# displaytext.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide classes which define binding and traversal methods for text.

The methods are shared by classes in cqldisplay and querydisplay modules
which display plain text.

The cqldisplay module has two sets of classes: based on the _CQLDisplay and
CQLDialogue classes.

The querydisplay module has an identical structure where the _CQLDisplay
and _QueryDisplay classes have many methods in common.

All methods in the classes in this module existed as multiple copies in various
classes in the cqldisplay and querydisplay modules.  They are now deleted
from those modules.

The classes in this module represent the different sets of classes with methods
in common.  Two classes could reasonably be called DisplayText: the choice is
consistent with the naming of InsertText and EditText.  ShowText was chosen for
the other class.

The ShowText class is populated with the methods identical in _CQLDisplay
and _QueryDisplay, which were then removed from those two classes.

The DisplayText class is populated with the methods identical in CQLDisplay
and QueryDisplay, which were then removed from those two classes.

The InsertText class is populated with the methods identical in
CQLDisplayInsert and QueryDisplayInsert, which were then removed from those
two classes.

The EditText class is populated with the methods identical in
CQLDisplayEdit and QueryDisplayEdit, which were then removed from those two
classes.  It probably should be a subclass of InsertText, but this depends on
successful choice of method resolution order in the classes in cqldisplay and
querydisplay modules.

"""
import tkinter
import tkinter.messagebox

from .eventspec import EventSpec
from .score import NonTagBind


# ShowText because DisplayText fits CQLDisplay (and *Query*), but there is
# no EngineDisplay class.  ScorePGN has no equivalent in the *Text line
# because selection rules, CQL statements, and run engine commands, do not
# have 'analysis'.
class ShowText:
    """Provide focus switching and visibility methods for text widgets."""

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    # Renamed from _bind_for_board_navigation to fit current use.
    # Before introduction of querydisplay and removal of board from predecessor
    # of cqldisplay on conversion, it was a reasonable name.
    def bind_for_item_navigation(self):
        """Set bindings to navigate text statement on pointer click."""
        self.set_score_pointer_item_navigation_bindings(True)

    def bind_for_widget_navigation(self):
        """Set pointer bindings to give focus to this text statement."""
        self.set_score_pointer_widget_navigation_bindings(True)

    def _get_close_item_events(self):
        """Return close item event description."""
        return ((EventSpec.display_dismiss, self.delete_item_view),)

    def set_and_tag_item_text(self, reset_undo=False):
        """Delegate to superclass method and set PGN score inactive."""
        # Superclass may set self._most_recent_bindings but test below must be
        # against current value.
        mrb = self._most_recent_bindings

        super().set_and_tag_item_text(reset_undo=reset_undo)
        if mrb != NonTagBind.NO_EDITABLE_TAGS:
            for event_spec in (self._get_inactive_button_events(),):
                self.set_event_bindings_score(event_spec, switch=True)

    def _create_database_submenu(self, menu):
        """Create and return popup submenu for database events."""
        submenu = tkinter.Menu(master=menu, tearoff=False)
        self._set_popup_bindings(submenu, self._get_database_events())
        return submenu

    # The only active bindings compared with displaypgn.ShowPGN.
    def _set_primary_activity_bindings(self, switch=True):
        """Delegate then set navigation and item close bindings."""
        super()._set_primary_activity_bindings(switch=switch)
        self._set_database_navigation_close_item_bindings(switch=switch)

    # Not relevant away from displaypgn.ShowPGN.
    # def _set_select_variation_bindings(self, switch=True):
    #    """Delegate to toggle other relevant bindings and toggle bindings for
    #    database actions, navigation to other widgets, and close widget.

    #    """
    #    super()._set_select_variation_bindings(switch=switch)
    #    self._set_database_navigation_close_item_bindings(switch=switch)

    # The methods identical except for docstrings, and references to
    # self.ui.game_items or self.ui.repertoire_items replaced by property
    # self.ui_displayed_items.

    def _next_item(self, event=None):
        """Select next item on display.

        Call _cycle_item after 1 millisecond to allow message display.

        """
        del event
        if self.ui_displayed_items.count_items_in_stack() > 1:
            self.ui_set_find_item_games(0)
            self.score.after(1, self.try_command(self._cycle_item, self.score))

    def _prior_item(self, event=None):
        """Select previous item on display.

        Call _cycle_item after 1 millisecond to allow message display.

        """
        del event
        if self.ui_displayed_items.count_items_in_stack() > 1:
            self.ui_set_find_item_games(-2)
            self.score.after(
                1, self.try_command(self._cycle_item, self.score), True
            )

    # What about _current_pgn_score call?
    def current_item(self, event=None):
        """Select current PGN score on display."""
        # cuiai should be referencing self given use of current_item() method,
        # but style of sibling *_item() methods is followed.
        # cuiai was cuigs in gamedisplay, and cuirs in repertoiredisplay,
        # modules originally.
        del event
        items = self.ui_displayed_items
        if items.count_items_in_stack():
            cuiai = items.active_item
            self._current_pgn_score(cuiai)
            cuiai.set_statusbar_text()

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order."""
        del event
        self.set_score_pointer_widget_navigation_bindings(True)
        self.ui.give_focus_backward(self.ui_displayed_items)
        return "break"

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order."""
        del event
        self.set_score_pointer_widget_navigation_bindings(True)
        self.ui.give_focus_forward(self.ui_displayed_items)
        return "break"

    # The methods identical except for docstrings, and references to
    # self.ui.configure_game_grid or self.ui.configure_repertoire_grid
    # replaced by property self.ui_configure_item_list_grid.

    def _cycle_item(self, prior=False):
        """Select next PGN score on display."""
        items = self.ui_displayed_items
        losefocus = items.active_item
        losefocus.bind_for_widget_navigation()
        items.cycle_active_item(prior=prior)
        self.ui_configure_item_list_grid()
        gainfocus = items.active_item
        gainfocus.set_game_list()
        gainfocus.bind_for_item_navigation()
        gainfocus.takefocus_widget.focus_set()
        gainfocus.set_statusbar_text()

    def give_focus_to_widget(self, event=None):
        """Select text item on display by mouse click."""
        self.ui.set_bindings_on_item_losing_focus_by_pointer_click()
        losefocus, gainfocus = self.ui_displayed_items.give_focus_to_widget(
            event.widget
        )
        if losefocus is not gainfocus:
            self.ui_configure_item_list_grid()
            self.score.after(
                0, func=self.try_command(self.ui_set_item_name, self.score)
            )
            self.score.after(
                0,
                func=self.try_command(gainfocus.refresh_game_list, self.score),
            )
        return "break"

    # The insert_game_database method, coerced into sameness from the methods
    # in gamedisplay._GameDisplay and repertoiredisplay._RepertoireDisplay with
    # class attibutes pgn_score_name, pgn_score_source, pgn_score_tags,
    # and method mark_partial_positions_to_be_recalculated, and property
    # ui_base_table.  The clarity of both common bits and differences
    # seems to justify the extra syntactic complexity.

    # Probably becomes _insert_item_database() in each subclass like in
    # cqldbshow and all other *db* calls like it.

    # This was not going to be moved to displaypgn except insert_game_database,
    # which uses the class attribute pgn_score_updater, was moved.
    # Both _game_updater and pgn_score_updater need more generic names.
    def _game_updater(self, text):
        """Make and return a chess record containing a single PGN score."""
        updater = self.pgn_score_updater()
        updater.value.load(text)
        return updater

    # Code shared by on_game_change method in _GameDisplay and
    # _RepertoireDisplay.
    # Replace _pgn_score_ with _item_ to fit with displaypgn.ShowPGN perhaps,
    # but this one will be moved to subclasses.
    def _patch_pgn_score_to_fit_record_change_and_refresh_grid(
        self, grid, instance
    ):
        """Adjust list of games same position if instance is for active game.

        A database update has been done for instance and if that game is the
        active game, the list of games matching the current position may be
        out of date.

        """
        if self.ui_displayed_items.is_item_panel_active(self):
            # Patch data structure to look as though the edited record has
            # been read from disk.  That means DataGrid, DisplayItems, and
            # this _GameDisplay or _RepertoireDisplay, instances.
            # self.sourceobject is DataGrid.get_visible_record(<key>), the
            # record prior to editing.
            # instance adds index structures to match those which should be
            # on database.
            # instance.newrecord is edited record including index stuff.
            key = None
            for grid_key in grid.keys:
                if instance.key.recno != grid_key[0]:
                    key = grid_key
                    break
            grid.close_client_cursor()
            grid.datasource.get_full_position_games(self._get_position_key())
            grid.fill_view(currentkey=key, exclude=False)


class DisplayText:
    """Provide method to set database insert and delete event descriptions."""

    def _get_database_events(self):
        """Return event description tuple for PGN score database actions."""
        return (
            (EventSpec.display_insert, self._insert_item_database),
            (EventSpec.display_delete, self._delete_item_database),
        )


class InsertText:
    """Provide method which generates database insert event descriptions."""

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    def _get_database_events(self):
        """Return event description tuple for PGN score database actions."""
        return ((EventSpec.display_insert, self._insert_item_database),)


# Introduced to remove _create_primary_activity_popup method from InsertText
# class: it's presence prevented displaypgn.InsertPGN being a subclass of
# InsertText.
# In the PGN classes the displayed list of games depends on which token is
# current, but in the non-PGN Text classes the displayed list of games is
# determined by evaluation of a query on demand.
# When inserting or editing a record the _create_primary_activity_popup method
# of this class makes an option to demand evaluation available in a popup menu.
# When showing a record the demand is implicit in making the 'show' widget
# active.
# In the PGN classes the demand is always implicit in making a token current,
# so there is no PGN equivalent of ListGamesText.
class ListGamesText:
    """Provide method which creates primary activity popup menu."""

    def _create_primary_activity_popup(self):
        """Delegate then add list games to popup and return popup menu."""
        popup = super()._create_primary_activity_popup()
        self._add_list_games_entry_to_popup(popup)
        return popup


class EditText:
    """Provide method which generates database update event descriptions."""

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    def _get_database_events(self):
        """Return event description tuple for PGN score database actions."""
        return (
            (EventSpec.display_insert, self._insert_item_database),
            (EventSpec.display_update, self._update_item_database),
        )

    # update_game_database becomes _update_item_database in cqldisplay and
    # querydisplay.



# read_pgn_profile.py
# Copyright 2018 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Read a PGN file and display timing profile."""

import cProfile
import pstats
import io
import tkinter

from pgn_read.samples import _utilities

from ..core.pgn import GameUpdate


class Main(_utilities.Main):
    """Collect timing profile for processing a PGN file."""

    def process_pgn_file(self, event=None):
        """Process PGN file."""
        pr = cProfile.Profile()
        pr.enable()
        super().process_pgn_file(event=event)
        pr.disable()
        s = io.StringIO()
        ps = pstats.Stats(pr, stream=s).sort_stats("cumulative")
        ps.print_stats()
        self.insert_text("\nProfile report\n\n")
        self.insert_text(s.getvalue())
        self.root.wm_resizable(width=tkinter.TRUE, height=tkinter.TRUE)
        self.root.columnconfigure(0, weight=1)
        self.text.pack(expand=tkinter.TRUE, fill=tkinter.BOTH)


if __name__ == "__main__":
    Main(
        game_class=GameUpdate,
        samples_title="Sample ChessTab Update Profile Report",
    ).root.mainloop()



# _lead_trail.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Class with attributes to number of leading and trailing spaces."""


class _LeadTrail:
    """Attributes for number of leading and trailing spaces."""

    def __init__(self, lead, trail):
        """Initialize the description."""
        self._lead = lead
        self._trail = trail

    @property
    def lead(self):
        """Return number of leading spaces."""
        return self._lead

    @lead.setter
    def lead(self, value):
        """Set self._lead to value."""
        self._lead = value

    @property
    def trail(self):
        """Return number of trailing spaces."""
        return self._trail

    @trail.setter
    def trail(self, value):
        """Set self._trail to value."""
        self._trail = value

    @property
    def header_length(self):
        """Return number of packing spaces associated with token."""
        return self._lead + self._trail



# sharedtext.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide classes which define navigation methods and bindings.

The classes are SharedText, SharedTextEngineText, and SharedTextScore.
"""

import tkinter

from .eventspec import EventSpec


class SharedText:
    """Provide navigation bindings for subclasses.

    Methods with the same name are used in one or both enginetext.EngineText
    and score.Score, but the internals are different.

    """

    def _set_primary_activity_bindings(self, switch=True):
        """Switch bindings for traversing query statement on or off."""
        self.set_event_bindings_score(
            self.get_f10_popup_events(
                self._post_active_menu_at_top_left, self._post_active_menu
            ),
            switch=switch,
        )
        self.set_event_bindings_score(
            self._get_button_events(buttonpress3=self._post_active_menu),
            switch=switch,
        )

    def set_score_pointer_item_navigation_bindings(self, switch):
        """Set or unset pointer bindings for game navigation."""
        self.set_event_bindings_score(
            self._get_button_events(buttonpress3=self._post_active_menu),
            switch=switch,
        )

    @staticmethod
    def get_primary_activity_events():
        """Return null tuple of navigation keypresses and callbacks."""
        return ()


class SharedTextEngineText:
    """Provide keypress and pointer navigation methods for subclasses.

    This is a separate hierarchy from SharedTextScore because the engine
    stuff is in a Toplevel instance not the main application widget.
    """

    @staticmethod
    def _get_modifier_buttonpress_suppression_events():
        """Return empty tuple of event binding definitions.

        These events suppress buttonpress with Control, Shift, or Alt.

        """
        return ()

    def _post_active_menu(self, event=None):
        """Show the popup menu for chess engine definition navigation."""
        return self._post_menu(
            self.primary_activity_popup,
            self._create_primary_activity_popup,
            allowed=self._is_active_item_mapped(),
            event=event,
        )

    def _post_active_menu_at_top_left(self, event=None):
        """Show the popup menu for chess engine definition navigation."""
        return self.post_menu_at_top_left(
            self.primary_activity_popup,
            self._create_primary_activity_popup,
            allowed=self._is_active_item_mapped(),
            event=event,
        )


class SharedTextScore:
    """Provide keypress and pointer navigation methods for subclasses.

    This is a separate hierarchy from SharedTextEngineText because the
    engine stuff is in a Toplevel instance not the main application
    widget.
    """

    def _add_cascade_menu_to_popup(
        self, label, popup, bindings=None, order=None, index=tkinter.END
    ):
        """Add cascade_menu, and bindings, to popup if not already present.

        The index is used as the label on the popup menu when visible.

        The bindings are not applied if cascade_menu is alreay in popup menu.

        """
        # Cannot see a way of asking 'Does entry exist?' other than:
        try:
            popup.index(label)
        except tkinter.TclError:
            cascade_menu = tkinter.Menu(master=popup, tearoff=False)
            popup.insert_cascade(label=label, menu=cascade_menu, index=index)
            if order is None:
                order = ()
            if bindings is None:
                bindings = {}
            for definition in order:
                function = bindings.get(definition)
                if function is not None:
                    cascade_menu.add_command(
                        label=definition[1],
                        command=self.try_command(function, cascade_menu),
                        accelerator=definition[2],
                    )

    # Subclasses which need widget navigation in their popup menus should
    # call this method.
    def _create_widget_navigation_submenu_for_popup(self, popup):
        """Create and populate a submenu of popup for widget navigation.

        The commands in the submenu should switch focus to another widget.

        Subclasses should define a generate_popup_navigation_maps method and
        binding_labels iterable suitable for allowed navigation.

        """
        navigation_map, local_map = self.generate_popup_navigation_maps()
        local_map.update(navigation_map)
        self._add_cascade_menu_to_popup(
            "Navigation", popup, bindings=local_map, order=self.binding_labels
        )

    # Subclasses which need dismiss widget in a menu should call this method.
    def _add_close_item_entry_to_popup(self, popup):
        """Add option to dismiss widget entry to popup.

        Subclasses must provide a delete_item_view method.

        """
        self._set_popup_bindings(popup, self._get_close_item_events())

    def _create_inactive_popup(self):
        """Create popup menu for an inactive widget."""
        assert self.inactive_popup is None
        popup = tkinter.Menu(master=self.score, tearoff=False)
        self._set_popup_bindings(popup, self._get_inactive_events())
        self._init_inactive_popup(popup)
        return popup

    def _get_inactive_button_events(self):
        """Return pointer event specifications for an inactive widget."""
        return self._get_modifier_buttonpress_suppression_events() + (
            (EventSpec.buttonpress_1, self.give_focus_to_widget),
            (EventSpec.buttonpress_3, self.post_inactive_menu),
        )

    def post_inactive_menu(self, event=None):
        """Show the popup menu for a game score in an inactive item."""
        return self._post_menu(
            self.inactive_popup, self._create_inactive_popup, event=event
        )

    def post_inactive_menu_at_top_left(self, event=None):
        """Show the popup menu for a game score in an inactive item."""
        return self.post_menu_at_top_left(
            self.inactive_popup, self._create_inactive_popup, event=event
        )

    def _get_inactive_events(self):
        """Return keypress event specifications for an inactive widget."""
        return (
            (EventSpec.display_make_active, self.set_focus_panel_item_command),
            (EventSpec.display_dismiss_inactive, self.delete_item_view),
        )



# cqlgames.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Evaluate a ChessQL query.

Methods defined on all database interfaces are used exclusively.

"""
from ast import literal_eval

from solentware_base.core.find import Find
from solentware_base.core.where import (
    Where,
    EQ,
    OR,
    NOT,
    AND,
)

from chessql.core import cql
from chessql.core.cql import Token
from chessql.core.constants import (
    ALL_GAMES_MATCH_PIECE_DESIGNATORS,
    ANY_WHITE_PIECE_NAME,
    ANY_BLACK_PIECE_NAME,
    ALL_PIECES,
    WHITE_PIECE_NAMES,
    BLACK_PIECE_NAMES,
    EMPTY_SQUARE_NAME,
)

from pgn_read.core.constants import (
    FEN_WHITE_KING,
)
from pgn_read.core.squares import Squares

from .rayfilter import move_number_str
from ..core.filespec import (
    PIECESQUAREMOVE_FIELD_DEF,
    PIECEMOVE_FIELD_DEF,
    SQUAREMOVE_FIELD_DEF,
    NEWGAMES_FIELD_DEF,
    PARTIAL_FILE_DEF,
    NEWGAMES_FIELD_VALUE,
    PARTIALPOSITION_FIELD_DEF,
)

# The immediate children of nodes with the names of these filters have 'and'
# and 'or' applied to evaluate the node's answer.
_AND_FILTERS = frozenset(
    (
        Token.AND,
        cql.LEFTBRACE_NUMBER,
        cql.LEFTBRACE_POSITION,
        cql.LEFTBRACE_SET,
        cql.LEFTBRACE_LOGICAL,
        Token.CQL,
    )
)
_OR_FILTERS = frozenset(
    (
        Token.OR,
        Token.FLIP,
        Token.FLIPHORIZONTAL,
        Token.FLIPVERTICAL,
        Token.FLIPCOLOR,
        Token.ROTATE45,
        Token.ROTATE90,
        Token.SHIFT,
        Token.SHIFTHORIZONTAL,
        Token.SHIFTVERTICAL,
    )
)

_filters = {
    Token.PIECE_DESIGNATOR,
    Token.RAY,
}
_SUPPORTED_FILTERS = frozenset(_AND_FILTERS.union(_OR_FILTERS.union(_filters)))
del _filters


class ChessQLGamesError(Exception):
    """Exception class for cqlgames module."""


class ChessQLGames:
    """Represent subset of games that match a Chess Query Language query."""

    def __init__(self, *a, **k):
        """Delete then create Find instance for queries."""
        super().__init__(*a, **k)

        # recordclass argument must be used to support non-index field finds.
        # Empty square searches can be transformed into equivalent occupied
        # square searches, but it does not work yet.
        # ChessDBrecordGameUpdate is the closest class to that needed.
        # No update capability is needed, and a different set of 'per move'
        # attributes might be better.
        # The 'try ...' statements in .gui.cqldisplay and .gui.cqltext fit the
        # version of Find(...) call with the recordclass argument.
        # self.cqlfinder = Find(
        #    self.dbhome, self.dbset, recordclass=ChessDBrecordGameUpdate)
        self.cqlfinder = Find(self.dbhome, self.dbset)

        self.not_implemented = set()

    def forget_cql_statement_games(self, sourceobject, commit=True):
        """Forget game records matching ChessQL statement.

        sourceobject is partial position record instance for cql query.
        commit indicates if this method should start and commit transaction.

        """
        if commit:
            self.dbhome.start_transaction()
        # Forget the list of games under the query key.
        ppview = self.dbhome.recordlist_record_number(
            PARTIAL_FILE_DEF, key=sourceobject.key.recno
        )
        self.dbhome.unfile_records_under(
            self.dbset,
            PARTIALPOSITION_FIELD_DEF,
            self.dbhome.encode_record_number(sourceobject.key.recno),
        )

        # Remove query from set that needs recalculating.
        changed = self.dbhome.recordlist_key(
            PARTIAL_FILE_DEF,
            NEWGAMES_FIELD_DEF,
            key=self.dbhome.encode_record_selector(NEWGAMES_FIELD_VALUE),
        )
        changed.remove_recordset(ppview)

        self.dbhome.file_records_under(
            PARTIAL_FILE_DEF,
            NEWGAMES_FIELD_DEF,
            changed,
            self.dbhome.encode_record_selector(NEWGAMES_FIELD_VALUE),
        )
        if commit:
            self.dbhome.commit()

    def update_cql_statement_games(
        self, sourceobject, initial_recordset=None, commit=True
    ):
        """Find game records matching ChessQL statement and store on database.

        sourceobject is partial position record instance for cql query.
        initial_recordset is games to which query is applied
        commit indicates if this method should start and commit transaction.

        """
        assert sourceobject is not None
        # Evaluate query.
        if commit:
            self.dbhome.start_read_only_transaction()
        if initial_recordset is None:
            initial_recordlist = self.dbhome.recordlist_ebm(self.dbset)
        else:
            initial_recordlist = self.dbhome.recordlist_nil(self.dbset)
            initial_recordlist |= initial_recordset
        query = sourceobject.value
        games = self._get_games_matching_filters(
            query,
            self._get_games_matching_parameters(query, initial_recordlist),
        )

        # File the list of games under the query key.
        ppview = self.dbhome.recordlist_record_number(
            PARTIAL_FILE_DEF, key=sourceobject.key.recno
        )
        # The read-only transaction is added for Symas LMMD.
        # The read-write transaction start is not moved to read-only start
        # to preserve the transaction boundaries in the other database
        # engines.  (Moving it should be ok though.)
        if commit:
            self.dbhome.end_read_only_transaction()
        if commit:
            self.dbhome.start_transaction()
        self.dbhome.file_records_under(
            self.dbset,
            PARTIALPOSITION_FIELD_DEF,
            games,
            self.dbhome.encode_record_number(sourceobject.key.recno),
        )

        # Remove query from set that needs recalculating.
        changed = self.dbhome.recordlist_key(
            PARTIAL_FILE_DEF,
            NEWGAMES_FIELD_DEF,
            key=self.dbhome.encode_record_selector(NEWGAMES_FIELD_VALUE),
        )
        changed.remove_recordset(ppview)

        self.dbhome.file_records_under(
            PARTIAL_FILE_DEF,
            NEWGAMES_FIELD_DEF,
            changed,
            self.dbhome.encode_record_selector(NEWGAMES_FIELD_VALUE),
        )
        if commit:
            self.dbhome.commit()

        # Hand the list of games over to the user interface.
        self.set_recordset(games)

    def get_cql_statement_games(
        self, query, sourceobject, initial_recordset=None, commit=True
    ):
        """Find game records matching ChessQL statement.

        query is detail extracted from query statement.
        sourceobject is previously calculated answer.  Set to None to force
        recalculation from query (after editing query statement usually).
        initial_recordset is games to which query is applied.

        """
        del commit
        if query is None:
            self.set_recordset(self.dbhome.recordlist_nil(self.dbset))
            return
        self.dbhome.start_read_only_transaction()

        # Use the previously calculated record set if possible.
        # sourceobject is set to None if query must be recalculated.
        if sourceobject is not None:
            ppview = self.dbhome.recordlist_record_number(
                PARTIAL_FILE_DEF, key=sourceobject.key.recno
            )
            pprecalc = self.dbhome.recordlist_key(
                PARTIAL_FILE_DEF,
                NEWGAMES_FIELD_DEF,
                key=self.dbhome.encode_record_selector(NEWGAMES_FIELD_VALUE),
            )
            pprecalc &= ppview
            if pprecalc.count_records() == 0:
                # At version 7.0.3.dev3 the workaround introduced at version
                # 7.0.3.dev1 was removed.
                # The problem was a missing record write in file_records_under
                # method in solentware_base's _lmdb module, not a transaction
                # design problem related to explicit read-only transactions,
                # as originally thought.
                ppcalc = self.dbhome.recordlist_key_startswith(
                    self.dbset,
                    PARTIALPOSITION_FIELD_DEF,
                    keystart=self.dbhome.encode_record_number(
                        sourceobject.key.recno
                    ),
                )

                if ppcalc.count_records() != 0:
                    games = self.dbhome.recordlist_key(
                        self.dbset,
                        PARTIALPOSITION_FIELD_DEF,
                        key=self.dbhome.encode_record_number(
                            sourceobject.key.recno
                        ),
                    )

                    # Hand the list of games over to the user interface.
                    self.set_recordset(games)

                    self.dbhome.end_read_only_transaction()
                    return

        # Evaluate query.
        if initial_recordset is None:
            initial_recordlist = self.dbhome.recordlist_ebm(self.dbset)
        else:
            initial_recordlist = self.dbhome.recordlist_nil(self.dbset)
            initial_recordlist |= initial_recordset
        games = self._get_games_matching_filters(
            query,
            self._get_games_matching_parameters(query, initial_recordlist),
        )
        self.dbhome.end_read_only_transaction()

        # Hand the list of games over to the user interface.
        self.set_recordset(games)

    # Not called anywhere.
    # Introduced at revision 3755 as part of 'on' filter implementation which
    # has been, at least temporarely, removed.
    # Called _pieces_matching_filter then.
    def pieces_matching_filter(self, filter_, initialpieces):
        """Return squares matching filters in CQL statement.

        It is assumed FSNode.designator_set contains the expanded piece
        designators.

        """
        if filter_.tokendef in _AND_FILTERS:
            pieces = set(initialpieces)
            for node in filter_.children:
                if node.tokendef is Token.PIECE_DESIGNATOR:
                    pieces.intersection_update(
                        self.pieces_matching_piece_designator(node)
                    )
                    if not pieces:
                        return pieces
            for node in filter_.children:
                if self._is_filter_not_implemented(node):
                    continue
                pieces.intersection_update(
                    self.pieces_matching_filter(node, pieces)
                )
                if not pieces:
                    return pieces
            return pieces
        if filter_.tokendef in _OR_FILTERS:
            pieces = set(ALL_PIECES)
            for node in filter_.children:
                if self._is_filter_not_implemented(node):
                    continue

                # This cannot be correct given pieces initialisation, or
                # more likely pieces initialisation is not correct too.
                # However pieces_matching_filter is not called anywhere!
                pieces.union(self.pieces_matching_filter(node, initialpieces))

            return pieces
        if filter_.tokendef is Token.PIECE_DESIGNATOR:
            return self.pieces_matching_piece_designator(filter_)
        return initialpieces

    # Not called anywhere.
    # Introduced at revision 3755 as part of 'on' filter implementation which
    # has been, at least temporarely, removed.
    # Called _squares_matching_filter then.
    # And back to _squares_matching_filter again! (revision 5067)
    def _squares_matching_filter(self, filter_, initialsquares):
        """Return squares matching filters in CQL statement.

        It is assumed FSNode.designator_set contains the expanded piece
        designators.

        """
        if filter_.tokendef in _AND_FILTERS:
            squares = set(initialsquares)
            for node in filter_.children:
                if node.tokendef is Token.PIECE_DESIGNATOR:
                    squares.intersection_update(
                        self.squares_matching_piece_designator(node)
                    )
                    if not squares:
                        return squares
            for node in filter_.children:
                if self._is_filter_not_implemented(node):
                    continue
                squares.intersection_update(
                    self._squares_matching_filter(node, squares)
                )
                if not squares:
                    return squares
            return squares
        if filter_.tokendef in _OR_FILTERS:
            squares = set(Squares.squares)
            for node in filter_.children:
                if self._is_filter_not_implemented(node):
                    continue

                # This cannot be correct given squares initialisation, or
                # more likely squares initialisation is not correct too.
                # However _squares_matching_filter is not called anywhere!
                squares.union(
                    self._squares_matching_filter(node, initialsquares)
                )

            return squares
        if filter_.tokendef is Token.PIECE_DESIGNATOR:
            return self.squares_matching_piece_designator(filter_)
        return initialsquares

    @staticmethod
    def pieces_matching_piece_designator(filter_):
        """Return pieces matching a piece designator."""
        if filter_.tokendef is not Token.PIECE_DESIGNATOR:
            raise ChessQLGamesError(
                "".join(
                    (
                        "'",
                        Token.PIECE_DESIGNATOR.name,
                        "' expected but '",
                        str(filter_.name),
                        "'received",
                    )
                )
            )
        pieces = set()
        for pdesignator in filter_.data.designator_set:
            piece = pdesignator[0]
            if piece == ANY_WHITE_PIECE_NAME:
                pieces.update(WHITE_PIECE_NAMES)
            elif piece == ANY_BLACK_PIECE_NAME:
                pieces.update(BLACK_PIECE_NAMES)
            else:
                pieces.add(piece)
        return pieces

    @staticmethod
    def squares_matching_piece_designator(filter_):
        """Return squares matching a piece designator."""
        if filter_.tokendef is not Token.PIECE_DESIGNATOR:
            raise ChessQLGamesError(
                "".join(
                    (
                        "'",
                        Token.PIECE_DESIGNATOR.name,
                        "' expected but '",
                        str(filter_.name),
                        "'received",
                    )
                )
            )
        squares = set()
        for pdesignator in filter_.data.designator_set:
            if len(pdesignator) == 1:
                squares.update(Squares.squares)
            else:
                squares.add(pdesignator[1:])
        return squares

    def _is_filter_not_implemented(self, filter_):
        """Return True if filter is not implemented, and False otherwise.

        Filters which are not implemented are added to a list reported before
        evaluating the statement to the extent possible.

        """
        if filter_.tokendef not in _SUPPORTED_FILTERS:
            self.not_implemented.add(filter_.tokendef.name)
            return True
        return False

    def _games_matching_piece_designator(
        self, filter_, datasource, movenumber, variation
    ):
        """Return games matching a piece designator."""
        if filter_.tokendef is not Token.PIECE_DESIGNATOR:
            raise ChessQLGamesError(
                "".join(
                    (
                        "'",
                        Token.PIECE_DESIGNATOR.name,
                        "' expected but '",
                        str(filter_.name),
                        "'received",
                    )
                )
            )
        answer = self._evaluate_piece_designator(
            movenumber, variation, filter_.data.designator_set
        )
        return datasource & answer

    def _games_matching_filter(
        self, filter_, initialgames, movenumber, variation
    ):
        """Return games matching filters in CQL statement.

        It is assumed FSNode.designator_set contains the expanded piece
        designators.

        """
        if filter_.tokendef in _AND_FILTERS:
            games = self.dbhome.recordlist_nil(self.dbset)
            games |= initialgames
            for node in filter_.children:
                if node.tokendef is Token.PIECE_DESIGNATOR:
                    games &= self._games_matching_piece_designator(
                        node, games, movenumber, variation
                    )
                    if not games.count_records():
                        return games
            for node in filter_.children:
                if self._is_filter_not_implemented(node):
                    continue
                games &= self._games_matching_filter(
                    node, games, movenumber, variation
                )
                if not games.count_records():
                    return games
            return games
        if filter_.tokendef in _OR_FILTERS:
            games = self.dbhome.recordlist_nil(self.dbset)
            for node in filter_.children:
                if self._is_filter_not_implemented(node):
                    continue
                games |= self._games_matching_filter(
                    node, initialgames, movenumber, variation
                )
            return games
        if filter_.tokendef is Token.PIECE_DESIGNATOR:
            return self._games_matching_piece_designator(
                filter_, initialgames, movenumber, variation
            )
        if filter_.tokendef is Token.RAY:
            return self._games_matching_ray_filter(
                filter_, initialgames, movenumber, variation
            )
        games = self.dbhome.recordlist_nil(self.dbset)
        games |= initialgames
        return games

    def _games_matching_ray_filter(
        self, filter_, datasource, movenumber, variation
    ):
        """Return games matching a ray filter."""
        del datasource
        if filter_.tokendef is not Token.RAY:
            raise ChessQLGamesError(
                "".join(
                    (
                        "'",
                        Token.RAY.name,
                        "' expected but '",
                        str(filter_.name),
                        "'received",
                    )
                )
            )
        ray_games = filter_.data.find_games(
            movenumber,
            variation,
            self._evaluate_ray_filter_element,
        )
        if ray_games:
            recordset = self.dbhome.recordlist_ebm(self.dbset)
            for value in ray_games.values():
                recordset &= value
        else:
            recordset = self.dbhome.recordlist_nil(self.dbset)
        return recordset

    def _evaluate_piece_designator(
        self, move_number, variation_code, piecesquares
    ):
        """Return games for piecesquares at move_number and variation_code.

        The return value is a RecordList instance.

        The move, variation, and piece designator need to be transformed
        into a value which potentially exists on the database.

        """
        return self._evaluate_statement(
            where_eq_piece_designator(
                move_number, variation_code, piecesquares
            )
        )

    def _evaluate_ray_filter_element(self, move_variation_square_piece):
        """Return games for move_variation_square_piece.

        The return value is a RecordList instance.

        The move, variation, square, and piece, have already been turned
        into a value which potentially exists on the database: no further
        transformation of the value is needed.

        """
        return self._evaluate_statement(
            where_eq_ray_element(move_variation_square_piece)
        )

    def _evaluate_statement(self, statement):
        """Return foundset calcualted from statement."""
        wqs = Where(statement)
        wqs.lex()
        wqs.parse()
        wqs.validate(self.cqlfinder.db, self.cqlfinder.dbset)
        wqs.evaluate(self.cqlfinder)
        return wqs.get_node_result_answer()

    def _get_games_matching_filters(self, query, games):
        """Select the games which meet the ChessQL cql() ... filters.

        Walk node tree for every movenumber and combine evaluated game sets.

        """
        found = self.dbhome.recordlist_nil(self.dbset)
        if not query.cql_filters:
            return found
        query.cql_filters.expand_set_filters()
        key = self._get_high_move_number()
        if key is None:
            return found
        self.not_implemented = set()
        for movenumber in move_number_in_key_range(key):
            found |= self._games_matching_filter(
                query.cql_filters, games, movenumber, "0"
            )
        return found

    # Version of _get_games_matching_filters() to process CQL parameters.
    def _get_games_matching_parameters(self, query, games):
        """Select the games which meet the ChessQL cql(...) parameters.

        Walk node tree for cql parameters and combine evaluated game sets.

        """
        del query
        found = self.dbhome.recordlist_nil(self.dbset)
        found |= games
        return found

    def _get_high_move_number(self):
        cursor = self.dbhome.database_cursor(
            self.dbset, PIECESQUAREMOVE_FIELD_DEF
        )
        try:
            return cursor.last()[0]
        except TypeError:
            return None
        finally:
            cursor.close()
            del cursor


def where_eq_piece_designator(move_number, variation_code, designator_set):
    """Return list of selection phrases to evaluate a piece designator.

    The fieldname PieceSquareMove is now unavoidably a bit confusing, where
    move refers to a token in the movetext when the piece is on the square.
    Nothing to do with move number in the movetext, which is the ChessQL
    meaning of move in this context.

    """
    anypiece = ANY_WHITE_PIECE_NAME + ANY_BLACK_PIECE_NAME
    psmfield = PIECESQUAREMOVE_FIELD_DEF
    pmfield = PIECEMOVE_FIELD_DEF
    smfield = SQUAREMOVE_FIELD_DEF
    mns = move_number_str(move_number)
    psmds = set()
    pmds = set()
    emptyds = set()
    smds = set()
    for pdesignator in designator_set:
        if len(pdesignator) == 1:
            # Rules of chess imply whole piece designator finds all games if
            # any of 'A', 'a', 'K', 'k', and '_', are in designator set.
            if pdesignator[0] in ALL_GAMES_MATCH_PIECE_DESIGNATORS:
                return " ".join(
                    (
                        pmfield,
                        EQ,
                        "".join((mns, variation_code, FEN_WHITE_KING)),
                    )
                )

            pmds.add("".join((mns, variation_code, pdesignator[0])))
            continue
        if pdesignator[0] == EMPTY_SQUARE_NAME:
            sqr = pdesignator[1:]
            emptyds.add(
                " ".join(
                    (
                        NOT,
                        smfield,
                        EQ,
                        "".join(
                            (mns, variation_code, sqr + ANY_WHITE_PIECE_NAME)
                        ),
                        AND,
                        NOT,
                        smfield,
                        EQ,
                        "".join(
                            (mns, variation_code, sqr + ANY_BLACK_PIECE_NAME)
                        ),
                        # Restrict sets generated by NOTs to games active at
                        # <mns, variation_code>.  'K' or 'k'  achieve this
                        # from the rules of chess.
                        AND,
                        pmfield,
                        EQ,
                        "".join((mns, variation_code, FEN_WHITE_KING)),
                    )
                )
            )
            continue
        if pdesignator[0] in anypiece:
            smds.add(
                "".join(
                    (
                        mns,
                        variation_code,
                        pdesignator[1:] + pdesignator[0],
                    )
                )
            )
            continue
        psmds.add(
            "".join(
                (
                    mns,
                    variation_code,
                    pdesignator[1:] + pdesignator[0],
                )
            )
        )

    phrases = []
    if psmds:
        phrases.append(" ".join((psmfield, EQ, OR.join("  ").join(psmds))))
    if pmds:
        phrases.append(" ".join((pmfield, EQ, OR.join("  ").join(pmds))))
    if smds:
        phrases.append(" ".join((smfield, EQ, OR.join("  ").join(smds))))
    if emptyds:
        phrases.append(OR.join("  ").join(emptyds))
    if phrases:
        return OR.join("  ").join(phrases)
    return None


def where_eq_ray_element(move_variation_square_piece):
    """Return list of selection phrases to evaluate a ray element.

    move_variation_square_piece is named for the order of the values of
    which it is composed.  THe value is to be placed in a 'where' clause.

    """
    anypiece = ANY_WHITE_PIECE_NAME + ANY_BLACK_PIECE_NAME
    psmfield = PIECESQUAREMOVE_FIELD_DEF
    pmfield = PIECEMOVE_FIELD_DEF
    smfield = SQUAREMOVE_FIELD_DEF
    psmds = set()
    emptyds = set()
    smds = set()
    if move_variation_square_piece[-1] == EMPTY_SQUARE_NAME:
        emptyds.add(
            " ".join(
                (
                    NOT,
                    smfield,
                    EQ,
                    move_variation_square_piece[:-1] + ANY_WHITE_PIECE_NAME,
                    AND,
                    NOT,
                    smfield,
                    EQ,
                    move_variation_square_piece[:-1] + ANY_BLACK_PIECE_NAME,
                    # Restrict sets generated by NOTs to games active at
                    # <mns, variation_code>.  'K' or 'k'  achieve this
                    # from the rules of chess.
                    AND,
                    pmfield,
                    EQ,
                    move_variation_square_piece[:-3] + FEN_WHITE_KING,
                )
            )
        )
    elif move_variation_square_piece[-1] in anypiece:
        smds.add(move_variation_square_piece)
    else:
        psmds.add(move_variation_square_piece)

    phrases = []
    if psmds:
        phrases.append(" ".join((psmfield, EQ, OR.join("  ").join(psmds))))
    if smds:
        phrases.append(" ".join((smfield, EQ, OR.join("  ").join(smds))))
    if emptyds:
        phrases.append(OR.join("  ").join(emptyds))
    if phrases:
        return OR.join("  ").join(phrases)
    return None


# def move_number_str(move_number):
#    """Return hex(move_number) values without leading '0x' but prefixed with
#    string length.
#    """
#    # Adapted from module pgn_read.core.parser method add_move_to_game().
#    try:
#        return MOVE_NUMBER_KEYS[move_number]
#    except IndexError:
#        c = hex(move_number)
#        return str(len(c)-2) + c[2:]


def move_number_in_key_range(key):
    """Yield the move number keys in a range one-by-one."""
    for number in range(1, literal_eval("0x" + key[1 : int(key[0]) + 1])):
        yield number



# filespec.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Files and fields for chess database.

Overrule field naming choices in core.filespec.FileSpec so DPT Fastload
can succeed.

DPT Fastload does not work with field names that are not upper case.

DPT restricts just the first character of field names to be upper case in
all other respects.

"""
from solentware_base.core.constants import (
    PRIMARY,
    SECONDARY,
    FIELDS,
)

from ..core import filespec


class FileSpec(filespec.FileSpec):
    """Extend to support DPT Fastload."""

    def __init__(self, **kargs):
        """Define chess database with upper case field names."""
        super().__init__(**kargs)
        for file in self.values():
            file[PRIMARY] = file[PRIMARY].upper()
            file[SECONDARY] = {
                key: key.upper() if value is None else value.upper()
                for key, value in file[SECONDARY].items()
            }
            file[FIELDS] = {
                key.upper(): value for key, value in file[FIELDS].items()
            }



# berkeleydbdu_dir.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import directory of PGN files with berkeleydb.chessdbdu to database."""


if __name__ == "__main__":
    from .directory_widget import DirectoryWidget
    from ..berkeleydb.chessberkeleydbdu import chess_database_du

    DirectoryWidget(chess_database_du, "berkeleydb")



# export_repertoire.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess repertoire exporters."""

from . import chessrecord, filespec
from .export_pgn_import_format import get_game_pgn_import_format

_ENCODING = "utf-8"


def export_all_repertoires_pgn(database, filename):
    """Export all repertoires in PGN export format."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordRepertoire()
    instance.set_database(database)
    cursor = database.database_cursor(
        filespec.REPERTOIRE_FILE_DEF, filespec.REPERTOIRE_FILE_DEF
    )
    try:
        with open(filename, "w", encoding=_ENCODING) as gamesout:
            current_record = cursor.first()
            while current_record:
                instance.load_record(current_record)
                if instance.value.collected_game.is_pgn_valid():
                    gamesout.write(
                        instance.value.collected_game.get_repertoire_pgn()
                    )
                    gamesout.write("\n\n")
                current_record = cursor.next()
    finally:
        cursor.close()
    return True


def export_all_repertoires_pgn_no_comments(database, filename):
    """Export all repertoires in PGN export format without comments."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordRepertoire()
    instance.set_database(database)
    cursor = database.database_cursor(
        filespec.REPERTOIRE_FILE_DEF, filespec.REPERTOIRE_FILE_DEF
    )
    try:
        with open(filename, "w", encoding=_ENCODING) as gamesout:
            current_record = cursor.first()
            while current_record:
                instance.load_record(current_record)
                game = (
                    instance.value.collected_game
                )  # pycodestyle line too long.
                if game.is_pgn_valid():
                    gamesout.write(game.get_repertoire_pgn_no_comments())
                    gamesout.write("\n\n")
                current_record = cursor.next()
    finally:
        cursor.close()
    return True


def export_all_repertoires_pgn_import_format(database, filename):
    """Export all repertoires in a PGN import format."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordRepertoire()
    instance.set_database(database)
    cursor = database.database_cursor(
        filespec.REPERTOIRE_FILE_DEF, filespec.REPERTOIRE_FILE_DEF
    )
    try:
        with open(filename, "w", encoding=_ENCODING) as gamesout:
            current_record = cursor.first()
            while current_record:
                instance.load_record(current_record)
                if instance.value.collected_game.is_pgn_valid():
                    gamesout.write(
                        get_game_pgn_import_format(
                            instance.value.collected_game
                        )
                    )
                    gamesout.write("\n\n")
                current_record = cursor.next()
    finally:
        cursor.close()
    return True


def export_all_repertoires_text(database, filename):
    """Export repertoires in database to text file in internal format."""
    if filename is None:
        return True
    instance = chessrecord.ChessDBrecordGameText()
    instance.set_database(database)
    cursor = database.database_cursor(
        filespec.REPERTOIRE_FILE_DEF, filespec.REPERTOIRE_FILE_DEF
    )
    try:
        with open(filename, "w", encoding=_ENCODING) as gamesout:
            current_record = cursor.first()
            while current_record:
                instance.load_record(current_record)
                gamesout.write(instance.get_srvalue())
                gamesout.write("\n")
                current_record = cursor.next()
    finally:
        cursor.close()
    return True


def export_selected_repertoires_pgn(grid, filename):
    """Export selected repertoires in PGN export format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    if grid.bookmarks:
        database = grid.get_data_source().dbhome
        instance = chessrecord.ChessDBrecordRepertoire()
        instance.set_database(database)
        with open(filename, "w", encoding=_ENCODING) as gamesout:
            for bookmark in sorted(grid.bookmarks):
                instance.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF, bookmark[0]
                    )
                )
                if instance.value.collected_game.is_pgn_valid_export_format():
                    gamesout.write(
                        instance.value.collected_game.get_repertoire_pgn()
                    )
        return True
    export_all_repertoires_pgn(grid.get_data_source().dbhome, filename)
    return True


def export_selected_repertoires_pgn_no_comments(grid, filename):
    """Export selected repertoires in PGN export format without comments.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return
    if grid.bookmarks:
        database = grid.get_data_source().dbhome
        instance = chessrecord.ChessDBrecordRepertoire()
        instance.set_database(database)
        with open(filename, "w", encoding=_ENCODING) as gamesout:
            for bookmark in sorted(grid.bookmarks):
                instance.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF, bookmark[0]
                    )
                )
                if instance.value.collected_game.is_pgn_valid_export_format():
                    rvcg = instance.value.collected_game  # For line length.
                    gamesout.write(rvcg.get_repertoire_pgn_no_comments())
        return
    export_all_repertoires_pgn_no_comments(
        grid.get_data_source().dbhome, filename
    )
    return


def export_selected_repertoires_pgn_import_format(grid, filename):
    """Export selected repertoires in a PGN import format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    instance = chessrecord.ChessDBrecordRepertoire()
    instance.set_database(database)
    games = []
    all_games_output = True
    if grid.bookmarks:
        for bookmark in grid.bookmarks:
            instance.load_record(
                database.get_primary_record(
                    filespec.REPERTOIRE_FILE_DEF, bookmark[0 if primary else 1]
                )
            )
            if instance.value.collected_game.is_pgn_valid_export_format():
                games.append(
                    get_game_pgn_import_format(instance.value.collected_game)
                )
            else:
                all_games_output = False
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                current_record = cursor.first()
            else:
                current_record = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while current_record:
                if not primary:
                    if not current_record[0].startswith(grid.partial):
                        break
                instance.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                if instance.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        get_game_pgn_import_format(
                            instance.value.collected_game
                        )
                    )
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            current_record = cursor.first()
            while True:
                if current_record is None:
                    break
                instance.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                if instance.value.collected_game.is_pgn_valid_export_format():
                    games.append(
                        get_game_pgn_import_format(
                            instance.value.collected_game
                        )
                    )
                else:
                    all_games_output = False
                current_record = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    with open(filename, "w", encoding=_ENCODING) as gamesout:
        for game in games:
            gamesout.write(game)
            gamesout.write("\n\n")
    return all_games_output


def export_selected_repertoires_text(grid, filename):
    """Export selected repertoires to text file in internal record format.

    If any records are bookmarked just the bookmarked records are exported,
    otherwise all records selected for display in the grid are exported.

    """
    if filename is None:
        return True
    database = grid.get_data_source().dbhome
    primary = database.is_primary(
        grid.get_data_source().dbset, grid.get_data_source().dbname
    )
    instance = chessrecord.ChessDBrecordRepertoire()
    instance.set_database(database)
    games = []
    if grid.bookmarks:
        for bookmark in grid.bookmarks:
            instance.load_record(
                database.get_primary_record(
                    filespec.REPERTOIRE_FILE_DEF, bookmark[0 if primary else 1]
                )
            )
            games.append(instance.get_srvalue())
    elif grid.partial:
        cursor = grid.get_cursor()
        try:
            if primary:
                current_record = cursor.first()
            else:
                current_record = cursor.nearest(
                    database.encode_record_selector(grid.partial)
                )
            while current_record:
                if not primary:
                    if not current_record[0].startswith(grid.partial):
                        break
                instance.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                games.append(instance.get_srvalue())
                current_record = cursor.next()
        finally:
            cursor.close()
    else:
        cursor = grid.get_cursor()
        try:
            current_record = cursor.first()
            while True:
                if current_record is None:
                    break
                instance.load_record(
                    database.get_primary_record(
                        filespec.REPERTOIRE_FILE_DEF,
                        current_record[0 if primary else 1],
                    )
                )
                games.append(instance.get_srvalue())
                current_record = cursor.next()
        finally:
            cursor.close()
    if len(games) == 0:
        return None
    with open(filename, "w", encoding=_ENCODING) as gamesout:
        for game in games:
            gamesout.write(game)
            gamesout.write("\n")
    return True


def export_single_repertoire_pgn(collected_game, filename):
    """Export repertoire in import format PGN to *.pgn file.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    with open(filename, "w", encoding=_ENCODING) as gamesout:
        gamesout.write(collected_game.get_repertoire_pgn())


def export_single_repertoire_pgn_no_comments(collected_game, filename):
    """Export repertoire in PGN export format without comments.

    Caller should test is_pgn_valid_export_format before picking filename.

    """
    if filename is None:
        return
    with open(filename, "w", encoding=_ENCODING) as gamesout:
        gamesout.write(collected_game.get_repertoire_pgn_no_comments())



# lmdbdu_file.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with lmdb.chessdbdu to database."""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..lmdb.chesslmdbdu import ChessDatabase

    FileWidget(ChessDatabase, "db")



# upgrade_3_to_4.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Upgrade a ChessTab database from version 3 to version 4.

Table names are changed and extra tables introduced in segment support.

"""


if __name__ == "__main__":
    import os

    from solentware_base.tools import ui_base_3_to_4

    from ..core.filespec import FileSpec

    ui_base_3_to_4.UIBase_3_to_4(FileSpec()).root.mainloop()



# gamerow.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Create widgets that display tag roster details of games on database."""

import tkinter

from solentware_grid.gui.datarow import (
    DataRow,
    GRID_COLUMNCONFIGURE,
    GRID_CONFIGURE,
    WIDGET_CONFIGURE,
    WIDGET,
    ROW,
)

from pgn_read.core.constants import (
    TAG_WHITE,
    TAG_BLACK,
    TAG_RESULT,
    TAG_EVENT,
    TAG_DATE,
    SEVEN_TAG_ROSTER,
    DEFAULT_TAG_VALUE,
    DEFAULT_TAG_DATE_VALUE,
    DEFAULT_TAG_RESULT_VALUE,
)

from ..core.chessrecord import ChessDBrecordGameTags
from . import constants
from ..shared.allrow import AllRow
from ..shared.game_position import GamePosition


class ChessDBrowGame(GamePosition, AllRow, ChessDBrecordGameTags, DataRow):
    """Define row in list of games.

    Add row methods to the chess game record definition.

    """

    header_specification = [
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_WHITE,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="player"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_RESULT,
                anchor=tkinter.CENTER,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=1, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="score"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_BLACK,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=2, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="player"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_EVENT,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=3, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="event"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_DATE,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=4, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="date"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text="Tags",
                anchor=tkinter.W,
                padx=10,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=5, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=4, uniform="tags"),
            ROW: 0,
        },
    ]

    def __init__(self, database=None, ui=None):
        """Extend and associate record definition with database.

        database - the open database that is source of row data
        ui - the ChessUI instamce

        """
        super().__init__()
        self.ui = ui
        self.set_database(database)
        self.row_specification = [
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.CENTER,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=1, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=2, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=3, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=4, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=10,
                ),
                GRID_CONFIGURE: dict(column=5, sticky=tkinter.EW),
                ROW: 0,
            },
        ]

    def grid_row(self, textitems=(), **kargs):
        """Set textitems to selected PGN tags, delegate, return response.

        Create textitems argument for ChessDBrowGame instance.

        textitems arguments is ignored and is present for compatibility.

        """
        tags = self.value.collected_game.pgn_tags
        return super().grid_row(
            textitems=(
                tags.get(TAG_WHITE, DEFAULT_TAG_VALUE),
                tags.get(TAG_RESULT, DEFAULT_TAG_RESULT_VALUE),
                tags.get(TAG_BLACK, DEFAULT_TAG_VALUE),
                tags.get(TAG_EVENT, DEFAULT_TAG_VALUE),
                tags.get(TAG_DATE, DEFAULT_TAG_DATE_VALUE),
                "  ".join(
                    [
                        "".join((tag, ' "', value, '"'))
                        for tag, value in self._get_tags_display_order(
                            self.value
                        )
                    ]
                ),
            ),
            **kargs
        )

    def _get_tags_display_order(self, pgn):
        """Return Tags not given their own column in display order."""
        del pgn
        str_tags = []
        other_tags = []
        tags = self.value.collected_game.pgn_tags
        for tag in SEVEN_TAG_ROSTER:
            if tag not in constants.GRID_HEADER_SEVEN_TAG_ROSTER:
                str_tags.append((tag, tags.get(tag, DEFAULT_TAG_VALUE)))
        for key, value in sorted(tags.items()):
            if key not in SEVEN_TAG_ROSTER:
                other_tags.append((key, value))
        return str_tags + other_tags


def chess_db_row_game(chessui):
    """Return function that returns ChessDBrowGame instance for chessui.

    chessui is a chess_ui.ChessUI instance.

    The returned function takes a ChessDatabase instance as it's argument.
    """

    def make_position(database=None):
        return ChessDBrowGame(database=database, ui=chessui)

    return make_position



# __init__.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access Berkeley DB database of chess games using berkeleydb package.

This interface can be used if the berkeleydb package is installed.

berkeleydb is available as a source package, from PyPI for example.

Follow the instructions to build and install berkeleydb.  This includes
installing Berkeley DB.
"""



# update_players_sqlite3.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Sample database traversal with update of Games file in Sqlite3 database.

Update names in games_Black and games_White records where the name with
',' and '.' ripped out is not the name in the record.
"""


if __name__ == "__main__":
    import os
    import ast
    import re
    import sqlite3

    re_normalize_player_name = re.compile(r"([^,\.\s]+(?:,|\.)?)(?:\s*)")

    table_names = "games_Black", "games_White"

    import tkinter.filedialog

    directoryname = tkinter.filedialog.askopenfilename(title="DB file to view")
    if directoryname:
        bn = os.path.basename(directoryname)
        sq = sqlite3.connect(directoryname)
        for t in table_names:
            print(t)
            select_statement = " ".join(("select", "*", "from", t))
            insert_statement = " ".join(
                (
                    "insert into",
                    t,
                    "(",
                    t,
                    ",",
                    "Segment",
                    ",",
                    "RecordCount",
                    ",",
                    "Game",
                    ")",
                    "values ( ? , ? , ? , ? )",
                )
            )
            delete_statement = " ".join(
                (
                    "delete from",
                    t,
                    "where",
                    t,
                    "== ? and",
                    "Segment",
                    "== ?",
                )
            )
            c = sq.cursor()
            d = sq.cursor()
            e = sq.cursor()
            c.execute(select_statement)
            while True:
                r = c.fetchone()
                if r is None:
                    break
                ks = r[0]
                nks = " ".join(re_normalize_player_name.findall(ks))
                if ks != nks:
                    # print(r)
                    try:
                        e.execute(insert_statement, (nks, r[1], r[2], r[3]))
                        d.execute(delete_statement, (ks, r[1]))
                    except sqlite3.IntegrityError:
                        print(r, nks)
                # print(r)
            c.close()
            d.close()
            e.close()
        sq.close()
        del sq



# dbdu.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""The Dbdu class for methods shared by Berkeley DB interface modules.

This module is relevant to the berkeleydb, bsddb3, and tcl, interfaces to
Berkeley DB.

"""
from ..core.filespec import (
    DB_ENVIRONMENT_GIGABYTES,
    DB_ENVIRONMENT_BYTES,
    DB_ENVIRONMENT_MAXLOCKS,
    DB_ENVIRONMENT_MAXOBJECTS,
)
from .litedu import Litedu


class Dbdu(Litedu):
    """Provide deferred update methods shared by the Berkeley DB interfaces.

    The methods provided by Litedu are shared with engines other than
    Berkeley DB.

    The whole database can be put in a single file, or each table (called a
    database in Berkeley DB terminology) in the database can be put in a
    file of it's own.
    """

    def __init__(self, databasefile, exception_class, flags, **kargs):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        environment = {
            "flags": flags,
            "gbytes": DB_ENVIRONMENT_GIGABYTES,
            "bytes": DB_ENVIRONMENT_BYTES,
            "maxlocks": DB_ENVIRONMENT_MAXLOCKS,
            "maxobjects": DB_ENVIRONMENT_MAXOBJECTS,
        }
        super().__init__(
            databasefile, exception_class, environment=environment, **kargs
        )



# cqledit.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Edit a Chess Query Language (ChessQL) statement.

ChessQL statements obey the syntax published for CQL version 6.0.1 (by Gady
Costeff).

The CQLEdit class extends the cql.CQL class to allow editing.

An instance of this class fits into the user interface in two ways: as an
item in a panedwindow of the main widget, or as the only item in a new toplevel
widget.

"""

from .cql import CQL


class CQLEdit(CQL):
    """Display a ChessQL statement with editing allowed.

    Attribute _is_text_editable is True meaning the statement can be
    edited.

    """

    # True means ChessQL statement can be edited
    _is_text_editable = True

    # Remove if this is all it is left doing.
    def __init__(self, **ka):
        """Extend ChessQL statement widget as editor."""
        super().__init__(**ka)



# querydbedit.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise edit toplevel to edit or insert game selection rule record."""
import tkinter.messagebox

from solentware_grid.gui.dataedit import DataEdit

from .querytoplevel import QueryToplevel, QueryToplevelEdit
from .topleveltext import EditText


class QueryDbEdit(EditText, DataEdit):
    """Edit game selection rule on database, or insert a new record.

    parent is used as the master argument in QueryToplevel calls.

    ui is used as the ui argument in QueryToplevel calls.

    newobject, parent, oldobject, and the one or two QueryToplevel instances
    created, are used as arguments in the super.__init__ call.

    showinitial determines whether a QueryToplevel is created for oldobject if
    there is one.

    Attribute text_name provides the name used in widget titles and message
    text.

    Methods _get_title_for_object and _set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    text_name = "Selection Rule Statement"

    def __init__(
        self,
        newobject=None,
        parent=None,
        oldobject=None,
        showinitial=True,
        ui=None,
    ):
        """Extend and create toplevel to edit or insert selection rule."""
        if not oldobject:
            showinitial = False
        super().__init__(
            newobject=newobject,
            parent=parent,
            oldobject=oldobject,
            newview=QueryToplevelEdit(master=parent, ui=ui),
            title="",
            oldview=QueryToplevel(master=parent, ui=ui)
            if showinitial
            else showinitial,
        )
        if ui is not None:
            nqs = self.newview.query_statement
            nqs.set_database(ui.base_games.datasource.dbhome)
            nqs.dbset = ui.base_games.datasource.dbset
            if showinitial:
                oqs = self.oldview.query_statement
                oqs.set_database(ui.base_games.datasource.dbhome)
                oqs.dbset = ui.base_games.datasource.dbset
        self._initialize()

    def _get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a selection rule object_.

        Default value of object_ is object attribute from DataShow class.

        """
        if object_ is None:
            object_ = self.oldobject
        if object_:
            return "  ".join(
                (
                    self.text_name.join(("Edit ", ":")),
                    object_.value.get_name_text(),
                )
            )
        return "".join(("Insert ", self.text_name))

    @property
    def ui_base_table(self):
        """Return the User Interface QueryGrid object."""
        return self.ui.base_selections

    @property
    def ui_items_in_toplevels(self):
        """Return the User Interface objects in Toplevels."""
        return self.ui.selections_in_toplevels

    @property
    def ui(self):
        """Return the User Interface object from 'editable' view."""
        return self.newview.ui

    @staticmethod
    def _set_item(view, object_):
        """Populate view with the query extracted from object_."""
        view.query_statement.process_query_statement(object_.get_srvalue())
        view.set_and_tag_item_text(reset_undo=True)

    def dialog_ok(self):
        """Return update action response (True for deleted).

        Delegate to superclass if the game selection rule is a valid statement
        or confirmation has been given for an invalid statement.

        """
        title = self._get_title_for_object()
        self.newobject.value.load(
            repr(self.newview.get_name_query_statement_text())
        )
        if not self.newobject.value.get_name_text():
            tkinter.messagebox.showerror(
                parent=self.parent,
                title=title,
                message="".join(
                    (
                        "The selection rule has no name.\n\nPlease enter ",
                        "it's name as the first line of text.'",
                    )
                ),
            )
            return False
        if self.newobject.value.where_error:
            if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.parent,
                title=title,
                message="".join(
                    (
                        "Confirm request to update game selection rule ",
                        "named:\n\n",
                        self.newobject.value.get_name_text(),
                        "\n\non database.\n\n",
                        self.newobject.value.where_error.get_error_report(
                            self.ui.base_games.get_data_source()
                        ),
                    )
                ),
            ):
                return False
        return super().dialog_ok()



# alldu.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Functions, and Alldu class for methods, used by most database interfaces.

'Most' means all database interfaces except DPT.
"""
import os
import traceback
import datetime

from solentware_base.core.segmentsize import SegmentSize
from solentware_base.core.constants import FILEDESC

from ..core.filespec import FileSpec
from ..core.chessrecord import ChessDBrecordGameImport
from .. import ERROR_LOG, APPLICATION_NAME


# This function is not absorbed in chess_du_import because the database
# has to be open for DPT backup, but not in the mode used to do an
# import.
def chess_du_backup_before_import(
    cdb,
    file=None,
    reporter=None,
    **kwargs,
):
    """Backup database cdb before import."""
    del kwargs
    if cdb.take_backup_before_deferred_update:
        if reporter is not None:
            reporter.append_text_only("")
            reporter.append_text("Make backup before import.")
        try:
            cdb.archive(name=file)
        except Exception as exc:
            _report_exception(cdb, reporter, exc)
            raise
        if reporter is not None:
            reporter.append_text("Backup completed.")


# This function is not absorbed in chess_du_import because in DPT the
# absence of 'file full' conditions has to be confirmed before deleting
# the backups.
def chess_du_delete_backup_after_import(
    cdb,
    file=None,
    reporter=None,
    **kwargs,
):
    """Delete backup of database cdb after import."""
    del kwargs
    if cdb.take_backup_before_deferred_update:
        if reporter is not None:
            reporter.append_text("Delete backup for import.")
        cdb.delete_archive(name=file)
        if reporter is not None:
            reporter.append_text("Backup deleted.")
            reporter.append_text_only("")
    if reporter is not None:
        reporter.append_text("Import finished.")
        reporter.append_text_only("")


def chess_du_import(
    cdb,
    pgnpaths,
    file=None,
    reporter=None,
    quit_event=None,
    increases=None,
):
    """Import games from pgnpaths into open database cdb."""
    importer = ChessDBrecordGameImport()
    for key in cdb.table.keys():
        if key == file:
            if increases is None:
                counts = [0, 0]
            else:
                counts = [increases[0], increases[1]]
            cdb.increase_database_record_capacity(files={key: counts})
            _chess_du_report_increases(reporter, key, counts)
            break
    else:
        if reporter is not None:
            reporter.append_text_only("")
            reporter.append_text(
                repr(file).join(
                    ("Unable to import to '", "': not found in database.")
                )
            )
            reporter.append_text_only("")
        return
    cdb.set_defer_update()
    try:
        for pgnfile in pgnpaths:
            with open(pgnfile, "r", encoding="iso-8859-1") as source:
                if not importer.import_pgn(
                    cdb,
                    source,
                    os.path.basename(pgnfile),
                    reporter=reporter,
                    quit_event=quit_event,
                ):
                    cdb.backout()
                    return
        if reporter is not None:
            reporter.append_text("Finishing import: please wait.")
            reporter.append_text_only("")
        cdb.do_final_segment_deferred_updates()
    except Exception as exc:
        _report_exception(cdb, reporter, exc)
        raise
    cdb.unset_defer_update()


def chess_du(cdb, *args, file=None, **kwargs):
    """Open database, delegate to chess_du_import, and close database."""
    chess_du_backup_before_import(cdb, file=file, **kwargs)
    cdb.open_database()
    chess_du_import(cdb, *args, file=file, **kwargs)
    cdb.close_database()
    chess_du_delete_backup_after_import(cdb, file=file, **kwargs)


def _chess_du_report_increases(reporter, file, size_increases):
    """Report size increases for file if any and there is a reporter.

    All elements of size_increases will be 0 (zero) if explicit increase
    in file size is not supported, or if not required when it is
    supported.

    """
    if reporter is None:
        return
    if sum(size_increases) == 0:
        return
    reporter.append_text_only("")
    reporter.append_text(file.join(("Increase size of '", "' file.")))
    label = ("Data", "Index")
    for item, size in enumerate(size_increases):
        reporter.append_text_only(
            " ".join(
                (
                    "Applied increase in",
                    label[item],
                    "pages:",
                    str(size),
                )
            )
        )


def _report_exception(cdb, reporter, exception):
    """Write exception to error log file, and reporter if available."""
    errorlog_written = True
    try:
        with open(
            os.path.join(cdb.home_directory, ERROR_LOG),
            "a",
            encoding="utf-8",
        ) as errorlog:
            errorlog.write(
                "".join(
                    (
                        "\n\n\n",
                        " ".join(
                            (
                                APPLICATION_NAME,
                                "exception report at",
                                datetime.datetime.isoformat(
                                    datetime.datetime.today()
                                ),
                            )
                        ),
                        "\n\n",
                        traceback.format_exc(),
                        "\n\n",
                    )
                )
            )
    except Exception:
        errorlog_written = False
    if reporter is not None:
        reporter.append_text("An exception has occured during import:")
        reporter.append_text_only("")
        reporter.append_text_only(str(exception))
        reporter.append_text_only("")
        if errorlog_written:
            reporter.append_text_only(
                "".join(
                    (
                        "detail appended to ",
                        os.path.join(cdb.home_directory, ERROR_LOG),
                        " file.",
                    )
                )
            )
        else:
            reporter.append_text_only(
                "".join(
                    (
                        "attempt to append detail to ",
                        os.path.join(cdb.home_directory, ERROR_LOG),
                        " file failed.",
                    )
                )
            )
        reporter.append_text_only("")
        reporter.append_text(
            "Import abandonned in way depending on database engine."
        )


def get_filespec(**kargs):
    """Return FileSpec instance with FILEDESCs removed at **kargs request.

    The FILEDESCs are deleted if allowcreate is False, the default.
    """
    names = FileSpec(**kargs)
    if not kargs.get("allowcreate", False):
        for table_name in names:
            if FILEDESC in names[table_name]:
                del names[table_name][FILEDESC]
    return names


class Alldu:
    """Provide deferred update methods shared by all interfaces.

    All the supported engines follow DPT in dividing the numeric primary
    keys into fixed-size segments.  When importing games a large amount of
    memory is required depending on number of games.  Some operating
    systems limit the memory available to a process.  The class attribute
    deferred_update_points is set when the segment size is greater than
    32768 in an attempt to avoid a MemoryError exception.
    """

    # The optimum chunk size is the segment size.
    # Assuming 2Gb memory:
    # A default FreeBSD user process will not cause a MemoryError exception for
    # segment sizes up to 65536 records, so the optimum chunk size defined in
    # the superclass will be selected.
    # A MS Windows XP process will cause the MemoryError exeption which selects
    # the 32768 game chunk size.
    # A default OpenBSD user process will cause the MemoryError exception which
    # selects the 16384 game chunk size.
    # The error log problem fixed at chesstab-0.41.9 obscured what was actually
    # happening: OpenBSD gives a MemoryError exception but MS Windows XP heads
    # for thrashing swap space in some runs with a 65536 chunk size (depending
    # on order of processing indexes I think). Windows 10 Task Manager display
    # made this obvious.
    # The MemoryError exception or swap space thrashing will likely not occur
    # for a default OpenBSD user process or a MS Windows XP process with
    # segment sizes up to 32768 records. Monitoring with Top and Task Manager
    # suggests it gets close with OpenBSD.

    # pylint comparison-with-callable report is false positive.
    # Perhaps because db_segment_size is a property and the last statement
    # in segmentsize module is 'SegmentSize = SegmentSize()'.
    if SegmentSize.db_segment_size > 32768:
        for f, m in ((4, 700000000), (2, 1400000000)):
            try:
                b" " * m
            except MemoryError:
                # Override the value in the superclass.
                deferred_update_points = frozenset(
                    i
                    for i in range(
                        65536 // f - 1,
                        SegmentSize.db_segment_size,
                        65536 // f,
                    )
                )

                break
        del f, m



# engine.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widget to display a chess engine definition.

A chess engine definition is the command to run a chess engine.

The Engine class displays a chess engine definition.

An instance of these classes fits into the user interface as the only item in a
new toplevel widget.

The engineedit module provides a subclass which allows editing the chess
engine definition.

The enginedbshow and enginedbedit modules provide subclasses used in a
new toplevel widget to display or edit chess engine definitions.

The enginedbdelete module provides a subclass used in a new toplevel widget
to allow deletion of chess engine definitions from a database.

"""

import tkinter

from .enginetext import EngineText


class Engine(EngineText):
    """Chess engine definition widget.

    master is used as the master argument for the tkinter Frame widget passed
    to superclass.

    boardfont is no longer used.

    See superclass for ui, items_manager, and itemgrid, arguments.  These may
    be, or have been, absorbed into **ka argument.

    """

    def __init__(
        self,
        master=None,
        boardfont=None,
        ui=None,
        items_manager=None,
        itemgrid=None,
        **ka
    ):
        """Create widgets to display chess engine definition."""
        del boardfont
        panel = tkinter.Frame(
            master=master, cnf=dict(borderwidth=2, relief=tkinter.RIDGE)
        )
        panel.grid_propagate(False)
        super().__init__(
            panel, ui=ui, items_manager=items_manager, itemgrid=itemgrid, **ka
        )
        self.scrollbar.grid(column=1, row=0, rowspan=1, sticky=tkinter.NSEW)
        self.score.grid(column=0, row=0, rowspan=1, sticky=tkinter.NSEW)
        if not ui.visible_scrollbars:
            panel.after_idle(self.hide_scrollbars)
        self.bind(
            panel, "<Configure>", function=self.try_event(self._on_configure)
        )
        self._configure_selection_widget()

        # For compatibility with Game when testing if item has focus.
        self.takefocus_widget = self.score

    def destroy_widget(self):
        """Destroy the widget displaying chess engine definition."""
        self.panel.destroy()

    def get_top_widget(self):
        """Return topmost widget for chess engine definition display.

        The topmost widget is put in a container widget in some way
        """
        return self.panel

    def _on_configure(self, event=None):
        """Reconfigure widget after container has been resized."""
        del event
        self._configure_selection_widget()

    def _configure_selection_widget(self):
        """Configure widgets for a chess engine definition display."""
        self.panel.grid_rowconfigure(0, weight=1)
        self.panel.grid_columnconfigure(0, weight=1)
        self.panel.grid_columnconfigure(1, weight=0)

    def hide_scrollbars(self):
        """Hide the scrollbars in chess engine definition display widgets."""
        self.scrollbar.grid_remove()
        self.score.grid_configure(columnspan=2)
        self._configure_selection_widget()

    def show_scrollbars(self):
        """Show the scrollbars in chess engine definition display widgets."""
        self.score.grid_configure(columnspan=1)
        self.scrollbar.grid_configure()
        self._configure_selection_widget()

    def takefocus(self, take=True):
        """Configure game widget takefocus option."""
        # Hack because I misunderstood meaning of takefocus: FALSE does not
        # stop the widget taking focus, just stops tab traversal.
        if take:
            # self.takefocus_widget.configure(takefocus=tkinter.TRUE)
            self.takefocus_widget.configure(takefocus=tkinter.FALSE)
        else:
            self.takefocus_widget.configure(takefocus=tkinter.FALSE)



# chessdptfastload.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using DPT fastload.

This module on Windows only.

See www.dptoolkit.com for details of DPT.

"""

import os
import multiprocessing  # Removed later by 'del multiprocessing'.

import tkinter
import tkinter.messagebox

# pylint will always give import-error message on non-Microsoft Windows
# systems.
# Wine counts as a Microsft Windows system.
# It is reasonable to not install 'dptdb.dptapi'.
# The importlib module is used to import chessdptdu if needed.
from dptdb.dptapi import FISTAT_DEFERRED_UPDATES

from solentware_base import dptfastload_database
from solentware_base.core.constants import (
    FILEDESC,
    BRECPPG,
    TABLE_B_SIZE,
)

from ..dpt.filespec import FileSpec
from ..core.filespec import (
    GAMES_FILE_DEF,
    PIECES_PER_POSITION,
    POSITIONS_PER_GAME,
)
from ..shared.alldu import chess_du_import

# The DPT segment size is 65280 because 32 bytes are reserved and 8160 bytes of
# the 8192 byte page are used for the bitmap.
# DB_SEGMENT_SIZE has no effect on processing apart from report points.
DB_SEGMENT_SIZE = 65280
_DEFERRED_UPDATE_POINTS = (DB_SEGMENT_SIZE - 1,)
del DB_SEGMENT_SIZE


class ChessdptfastloadError(Exception):
    """Exception class for chessdptfastload module."""


def chess_dptfastload(
    dbpath, pgnpaths, file_records=None, reporter=None, quit_event=None
):
    """Open database, import games and close database."""
    cdb = ChessDatabase(dbpath, allowcreate=True)
    cdb.open_database(files=file_records)

    # Intend to start a process, via multiprocessing, to do the database
    # update.  That process will do the reporting, not the one running
    # this method.
    chess_du_import(cdb, pgnpaths, reporter=reporter, quit_event=quit_event)

    cdb.close_database_contexts(files=file_records)
    cdb.open_database_contexts(files=file_records)
    status = True
    for file in (
        cdb.specification.keys() if file_records is None else file_records
    ):
        if cdb.table[file].get_file_parameters(cdb.dbenv)["FISTAT"][0]:
            status = False
    cdb.close_database_contexts()
    return status


class ChessDatabase(dptfastload_database.Database):
    """Provide fast load deferred methods for a database of games of chess.

    Subclasses must include a subclass of dptbase.Database as a superclass.

    """

    # deferred_update_points is the existing name but here it signifies
    # the interval at which a fastload call updates the database.
    # The number is arbitrary in relation to segment sizes since there may
    # be gaps in the record number sequence due to long games preventing
    # all slots in a page being used.
    deferred_update_points = frozenset(_DEFERRED_UPDATE_POINTS)

    # Override
    # dptfastload_database.Database._take_backup_before_deferred_update
    # while DPT archive() and delete_archive() methods absent.
    _take_backup_before_deferred_update = False

    def __init__(
        self,
        databasefolder,
        use_specification_items=None,
        dpt_records=None,
        **kargs,
    ):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        ddnames = FileSpec(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
        )
        # Deferred update for games file only
        for ddname in list(ddnames.keys()):
            if ddname != GAMES_FILE_DEF:
                del ddnames[ddname]

        if not kargs.get("allowcreate", False):
            try:
                for ddname in ddnames:
                    if FILEDESC in ddnames[ddname]:
                        del ddnames[ddname][FILEDESC]
            except Exception as error:
                if __name__ == "__main__":
                    raise
                raise ChessdptfastloadError(
                    "DPT description invalid"
                ) from error

        try:
            super().__init__(ddnames, databasefolder, **kargs)
        except ChessdptfastloadError as error:
            if __name__ == "__main__":
                raise
            raise ChessdptfastloadError("DPT description invalid") from error

        # Retain import estimates for increase size by button actions
        self._import_estimates = None
        self._notional_record_counts = None
        # Methods passed by UI to populate report widgets
        self._reporter = None

    def open_database(self, files=None):
        """Delegate then return None if database in deferred update mode.

        Close the database and raise ChessdptfastloadError exception if the
        database FISTAT parameter is not equal FISTAT_DEFERRED_UPDATES.

        """
        super().open_database(files=files)
        viewer = self.dbenv.Core().GetViewerResetter()
        for dbo in self.table.values():
            if viewer.ViewAsInt("FISTAT", dbo.opencontext):
                break
        else:
            if files is None:
                files = dict()
            self.increase_database_size(files=files)
            return
        self.close_database()
        raise ChessdptfastloadError("A file is not in deferred update mode")

    def open_context_prepare_import(self):
        """Open all files normally."""
        super().open_database()

    def get_pages_for_record_counts(self, counts=(0, 0)):
        """Return Table B and Table D pages needed for record counts."""
        brecppg = self.table[GAMES_FILE_DEF].filedesc[BRECPPG]
        return (
            counts[0] // brecppg,
            (counts[1] * self.table[GAMES_FILE_DEF].btod_factor) // brecppg,
        )

    def _get_database_table_sizes(self, files=None):
        """Return Table B and D size and usage in pages for files."""
        if files is None:
            files = dict()
        filesize = dict()
        for key, value in self.get_database_parameters(
            files=list(files.keys())
        ).items():
            filesize[key] = (
                value["BSIZE"],
                value["BHIGHPG"],
                value["DSIZE"],
                value["DPGSUSED"],
            )
        increase = self.get_database_increase(files=files)
        self.close_database_contexts()
        return filesize, increase

    def add_import_buttons(
        self,
        master,
        try_command_wrapper,
        try_event_wrapper,
        bind,
        widget,
        *args,
    ):
        """Add button actions for DPT to Import dialogue.

        Increase data and index space available.

        """
        index = tkinter.Button(
            master=master,
            text="Increase Index",
            underline=13,
            command=try_command_wrapper(self._increase_index, master),
        )
        index.pack(side=tkinter.RIGHT, padx=12)
        bind(widget, "<Alt-x>", try_event_wrapper(self._increase_index))
        data = tkinter.Button(
            master=master,
            text="Increase Data",
            underline=9,
            command=try_command_wrapper(self._increase_data, master),
        )
        data.pack(side=tkinter.RIGHT, padx=12)
        bind(widget, "<Alt-d>", try_event_wrapper(self._increase_data))

    def _increase_data(self, event=None):
        """Add maximum of current free space and default size to Table B.

        event is ignored and is present for compatibility between button click
        and keypress,

        """
        del event
        self.open_database_contexts(files=(GAMES_FILE_DEF,))
        increase_done = False
        for key, value in self.get_database_parameters(
            files=(GAMES_FILE_DEF,)
        ).items():
            bsize = value["BSIZE"]
            bused = max(0, value["BHIGHPG"])
            bneeded = self.get_pages_for_record_counts(
                self._notional_record_counts[key]
            )[0]
            bincrease = min(bneeded * 2, bsize - bused)
            message = "".join(
                (
                    "The free data size of the ",
                    key,
                    " file will be increased from ",
                    str(bsize - bused),
                    " pages to ",
                    str(bincrease + bsize - bused),
                    " pages.",
                )
            )
            if len(self.table[key].get_extents()) % 2 == 0:
                message = "".join(
                    (
                        message,
                        "\n\nAt present it is better to do index increases ",
                        "first for this file, if you need to do any, because ",
                        "a new extent (fragment) would not be needed.",
                    )
                )
            if tkinter.messagebox.askyesno(
                title="Increase Data Size",
                message="".join(
                    (
                        message,
                        "\n\nDo you want to increase the data size?",
                    )
                ),
            ):
                increase_done = True
                self.table[key].opencontext.Increase(bincrease, False)
        if increase_done:
            self._reporter.append_text(
                " ".join(
                    (
                        "Recalculation of planned database size increases",
                        "after data size increase by user action.",
                    )
                )
            )
            self._reporter.append_text_only("")
            self._report_plans_for_estimate()
        self.close_database_contexts()

    def _increase_index(self, event=None):
        """Add maximum of current free space and default size to Table D.

        event is ignored and is present for compatibility between button click
        and keypress,

        """
        del event
        self.open_database_contexts(files=(GAMES_FILE_DEF,))
        increase_done = False
        for key, value in self.get_database_parameters(
            files=(GAMES_FILE_DEF,)
        ).items():
            dsize = value["DSIZE"]
            dused = value["DPGSUSED"]
            dneeded = self.get_pages_for_record_counts(
                self._notional_record_counts[key]
            )[1]
            dincrease = min(dneeded * 2, dsize - dused)
            message = "".join(
                (
                    "The free index size of the ",
                    key,
                    " file will be increased from ",
                    str(dsize - dused),
                    " pages to ",
                    str(dincrease + dsize - dused),
                    " pages.",
                )
            )
            if len(self.table[key].get_extents()) % 2 != 0:
                message = "".join(
                    (
                        message,
                        "\n\nAt present it is better to do data increases ",
                        "first for this file, if you need to do any, because ",
                        "a new extent (fragment) would not be needed.",
                    )
                )
            if tkinter.messagebox.askyesno(
                title="Increase Index Size",
                message="".join(
                    (
                        message,
                        "\n\nDo you want to increase the index size?",
                    )
                ),
            ):
                increase_done = True
                self.table[key].opencontext.Increase(dincrease, True)
        if increase_done:
            self._reporter.append_text(
                " ".join(
                    (
                        "Recalculation of planned database size increases",
                        "after index size increase by user action.",
                    )
                )
            )
            self._reporter.append_text_only("")
            self._report_plans_for_estimate()
        self.close_database_contexts()

    def get_file_sizes(self):
        """Return dictionary of notional record counts for data and index."""
        return self._notional_record_counts

    def report_plans_for_estimate(self, estimates, reporter):
        """Calculate and report file size adjustments to do import.

        Note the reporter and headline methods for initial report and possible
        later recalculations.

        Pass estimates through to self._report_plans_for_estimate

        """
        # See comment near end of class definition Chess in relative module
        # ..gui.chess for explanation of this change.
        self._reporter = reporter
        self._report_plans_for_estimate(estimates=estimates)

    def _report_plans_for_estimate(self, estimates=None):
        """Recalculate and report file size adjustments to do import.

        Create dictionary of effective game counts for sizing Games file.
        This will be passed to the import job which will increase Table B and
        Table D according to file specification.

        The counts for Table B and Table D can be different.  If the average
        data bytes per game is greater than Page size / Records per page the
        count must be increased to allow for the unused record numbers.  If
        the average positions per game or pieces per position are not the
        values used to calculate the steady-state ratio of Table B to Table D
        the count must be adjusted to compensate.

        """
        append_text = self._reporter.append_text
        append_text_only = self._reporter.append_text_only
        if estimates is not None:
            self._import_estimates = estimates
        (
            gamecount,
            bytes_per_game,
            positions_per_game,
            pieces_per_game,
        ) = self._import_estimates[:4]
        brecppg = self.table[GAMES_FILE_DEF].filedesc[BRECPPG]
        d_count = (gamecount * (positions_per_game + pieces_per_game)) // (
            POSITIONS_PER_GAME * (1 + PIECES_PER_POSITION)
        )
        if bytes_per_game > (TABLE_B_SIZE // brecppg):
            b_count = int(
                (gamecount * bytes_per_game) / (TABLE_B_SIZE / brecppg)
            )
        else:
            b_count = gamecount
        self._notional_record_counts = {
            GAMES_FILE_DEF: (b_count, d_count),
        }
        append_text("Current file size and free space:")
        free = dict()
        sizes, increases = self._get_database_table_sizes(
            files=self._notional_record_counts
        )
        for filename, bdsize in sizes.items():
            bsize, bused, dsize, dused = bdsize
            bused = max(0, bused)
            free[filename] = (bsize - bused, dsize - dused)
            append_text_only(filename)
            append_text_only(
                " ".join(("Current data area size", str(bsize), "pages"))
            )
            append_text_only(
                " ".join(("Current index area size", str(dsize), "pages"))
            )
            append_text_only(
                " ".join(
                    ("Current data area free", str(bsize - bused), "pages")
                )
            )
            append_text_only(
                " ".join(
                    ("Current index area free", str(dsize - dused), "pages")
                )
            )
        append_text_only("")
        append_text("File space needed for import:")
        for filename, nr_count in self._notional_record_counts.items():
            append_text_only(filename)
            b_pages, d_pages = self.get_pages_for_record_counts(nr_count)
            append_text_only(
                " ".join(("Estimated", str(b_pages), "pages needed for data"))
            )
            append_text_only(
                " ".join(
                    ("Estimated", str(d_pages), "pages needed for indexes")
                )
            )
        append_text_only("")
        append_text("File size increases planned and free space when done:")
        for filename, increments in increases.items():
            b_incr, d_incr = increments
            b_free, d_free = free[filename]
            append_text_only(filename)
            append_text_only(
                " ".join(("Data area increase", str(b_incr), "pages"))
            )
            append_text_only(
                " ".join(("Index area increase", str(d_incr), "pages"))
            )
            append_text_only(
                " ".join(("Data area free", str(b_incr + b_free), "pages"))
            )
            append_text_only(
                " ".join(("Index area free", str(d_incr + d_free), "pages"))
            )
        append_text_only("")
        append_text_only(
            "".join(
                (
                    "Comparison of the required and free data or index ",
                    "space may justify using the Increase Data and, or, ",
                    "Increase Index actions to get more space immediately ",
                    "given your knowledge of the PGN file being imported.",
                )
            )
        )
        append_text_only("")
        append_text_only("")



# file_widget.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""The widget to drive import of a PGN file.

This module contains FileWidget, the code common to the original *du_file
modules; and file_du, a cut-down version of the chess_*du functions in the
engine specific chess*du modules (db.chessdbdu for example).
"""

import tkinter
import os

import tkinter.messagebox
import tkinter.filedialog

from ..core.chessrecord import ChessDBrecordGameImport
from ..core.filespec import GAMES_FILE_DEF


def file_du(database, dbpath, pgnpath, **kwargs):
    """Open database, import games and close database."""
    # The name argument in the archive() and delete_archive() calls is
    # needed if 'file_per_database=True' is in **kwargs, it is ignored
    # otherwise.
    cdb = database(dbpath, allowcreate=True, **kwargs)
    importer = ChessDBrecordGameImport()
    cdb.open_database()
    if cdb.take_backup_before_deferred_update:
        cdb.archive(name=GAMES_FILE_DEF)
    cdb.set_defer_update()
    s = open(pgnpath, "r", encoding="iso-8859-1")
    importer.import_pgn(cdb, s, os.path.basename(pgnpath))
    s.close()
    cdb.do_final_segment_deferred_updates()
    cdb.unset_defer_update()
    if cdb.take_backup_before_deferred_update:
        cdb.delete_archive(name=GAMES_FILE_DEF)
    cdb.close_database()


class FileWidget:
    """Provide select PGN game file dialogue and import from selected file."""

    def __init__(self, database, engine_name, **kwargs):
        """Import games into database using engine_name database engine."""
        root = tkinter.Tk()
        root.wm_title(string=" - ".join((engine_name, "Import PGN file")))
        root.wm_iconify()
        dbdir = tkinter.filedialog.askdirectory(
            title=" - ".join((engine_name, "Open ChessTab database"))
        )
        if dbdir:
            filename = tkinter.filedialog.askopenfilename(
                title="PGN file of Games",
                defaultextension=".pgn",
                filetypes=(("PGN Chess Games", "*.pgn"),),
            )
            if filename:
                if tkinter.messagebox.askyesno(
                    title="Import Games", message="Proceed with import"
                ):
                    file_du(database, dbdir, filename, **kwargs)
        root.destroy()



# displayitems.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Panel manager for sets of displayed items.

The classes for individual games, and so forth, expect their widgets to be
displayed in some container (panel).  In some circumstances it is convenient
to use the DisplayItemsStub class to give appropriate answers: when the
container is not used.

There are four kinds of item: game, repertoire, partial position, and selection
rule.

"""

import tkinter

# To support workaround for:
# _tkinter.TclError: window id "<number>" doesn't exist in this application
# which occurs on 64-bit Windows 7 Home Edition running 64-bit Python 3.4.3
# but not on 32-bit Windows XP.
# Not yet known if 64-bit OSes in general are affected.
# The problem code is: widget.winfo_pathname(widget.winfo_id())
import sys
import os

_WIN32_PLATFORM = sys.platform == "win32"
del sys

_amd64 = (
    os.getenv("PROCESSOR_ARCHITECTURE") == "AMD64"
    or os.getenv("PROCESSOR_ARCHITEW6432") == "AMD64"
)
del os


class DisplayItemsError(Exception):
    """Exception class for displayitems module."""


class DisplayItems:
    """Manage set of displayed widgets."""

    def __init__(self):
        """Create control data structures for  widgets."""
        self.order = []  # items in top to bottom display order.
        self.stack = []  # items in most recent visit order.
        self.panel_object_map = {}  # map panel identity to object displayed
        self.object_panel_count = {}  # count panels displaying an object

    def add_item_to_display(self, item):
        """Add item widget to GUI."""
        stack = self.stack
        order = self.order
        if order:
            order.insert(order.index(stack[-1]) + 1, item)
            stack.insert(0, item)
        else:
            stack.append(item)
            order.append(item)

    def contains_one_item(self):
        """Return True if order contains one element."""
        return len(self.order) == 1

    def get_active_item_top_widget(self):
        """Return top widget of item."""
        return self.stack[-1].get_top_widget()

    def is_item_panel_active(self, item):
        """Return True if itempanel's item is the active item."""
        if item.panel in self.panel_object_map:
            if item is self.stack[-1]:
                return True
        return None

    def is_mapped_panel(self, panel):
        """Return True if panel is in self.panel_object_map."""
        return panel in self.panel_object_map

    def is_visible(self):
        """Return True if active item exists."""
        return bool(self.stack)

    def count_items_in_stack(self):
        """Return number of items in stack.

        May be not equal count of items in panel_object_map or objects if item
        has been created but add_item_to_display() has not yet been called.

        """
        return len(self.stack)

    @property
    def active_item(self):
        """Return the active item."""
        try:
            return self.stack[-1]
        except IndexError:
            if self.stack:
                raise
            return None

    def get_stack_item(self, index):
        """Return self.stack[index]."""
        return self.stack[index]

    def cycle_active_item(self, prior=False):
        """Make active an item adjacent to current active item."""
        stack = self.stack
        order = self.order
        if prior:
            index = order.index(stack[-1]) - 1
            if index < 0:
                index = len(order) - 1
        else:
            index = order.index(stack[-1]) + 1
            if index >= len(order):
                index = 0
        stack.append(stack.pop(stack.index(order[index])))

    def any_items_displayed_of_type(self, class_=None):
        """Return True if instances of class_ displayed, default any class."""
        if class_ is None:
            return bool(self.stack)
        for item in self.stack:
            if isinstance(item, class_):
                return True
        return False

    def delete_item(self, item):
        """Delete item and return True if it was the active item."""
        stack = self.stack
        index = stack.index(item)
        stack[index].ui.set_game_change_notifications(stack[index])
        stack[index].destroy_widget()
        del self.order[self.order.index(stack[index])]
        del stack[index]

        # stack[-1] is always the active item.
        return index == len(stack)

    def delete_item_counters(self, item):
        """Delete panel, decrement counters, and return grid key for reset."""
        key = self.panel_object_map.get(item.panel, None)
        if key:
            del self.panel_object_map[item.panel]
        if key in self.object_panel_count:
            self.object_panel_count[key] -= 1
            if self.object_panel_count[key] == 0:
                del self.object_panel_count[key]
                return key
        return None

    def set_itemmap(self, item, objectkey):
        """Set panel_object_map to map item.panel to objectkey (database key).

        Panel is a surrogate for item in this map because item cannot be key
        in a dictionary,  Each item has one panel which contains everything.

        """
        self.panel_object_map[item.panel] = objectkey

    def increment_object_count(self, key):
        """Increment objects[key] to count widgets displaying object."""
        self.object_panel_count[key] = self.object_panel_count.get(key, 0) + 1

    def decrement_object_count(self, key):
        """Decrement objects[key] to count widgets displaying object."""
        self.object_panel_count[key] = self.object_panel_count.get(key, 0) - 1
        if self.object_panel_count[key] < 1:
            del self.object_panel_count[key]

    def adjust_edited_item(self, updater):
        """Fit self and active item state to database edit from updater values.

        It is assumed caller has checked it is the active item, or that only
        the active item gets to a point where it call this method.

        """
        item = self.stack[-1]
        if item.blockchange is not True:
            return False
        panel = item.panel
        pom = self.panel_object_map
        if panel not in pom:
            return False
        oldkey = pom[panel]
        recno = updater.key.recno
        if oldkey[0] != recno:
            return False
        opc = self.object_panel_count
        opc[oldkey] -= 1
        if opc[oldkey] <= 0:
            del opc[oldkey]
        newkey = recno, updater.srvalue
        opc[newkey] = opc.get(newkey, 0) + 1
        pom[panel] = newkey
        item.sourceobject.srvalue = updater.srvalue
        item.blockchange = False
        return newkey

    def give_focus_to_widget(self, widget):
        """Give focus to widget and return (lose focus, gain focus) widgets."""
        stack = self.stack
        losefocus = stack[-1]

        # win32 amd64 workaround. See comment at top of module.
        # Code in except clause is semantically closer to try clause than
        # a str(widget) version.
        try:
            gain = widget.winfo_pathname(widget.winfo_id())
        except tkinter.TclError:
            if not _WIN32_PLATFORM or not _amd64:
                raise
            gain = ".".join((widget.winfo_parent(), widget.winfo_name()))

        for item in stack:
            top_widget = item.get_top_widget()

            # win32 amd64 workaround. See comment at top of module.
            # Code in except clause is semantically closer to try clause than
            # a str(widget) version.
            try:
                if gain.startswith(
                    top_widget.winfo_pathname(top_widget.winfo_id())
                ):
                    gainfocus = item
                    break
            except tkinter.TclError:
                if not _WIN32_PLATFORM or not _amd64:
                    raise
                if gain.startswith(
                    ".".join(
                        (top_widget.winfo_parent(), top_widget.winfo_name())
                    )
                ):
                    gainfocus = item
                    break

        else:
            gainfocus = losefocus
        self.stack[-1].ui.set_toolbarframe_disabled()
        if losefocus is not gainfocus:
            stack.append(stack.pop(stack.index(gainfocus)))
            losefocus.bind_for_widget_navigation()
        gainfocus.bind_for_item_navigation()
        gainfocus.takefocus_widget.focus_set()
        if gainfocus.ui.single_view:
            gainfocus.ui.show_just_panedwindow_with_focus(
                gainfocus.get_top_widget()
            )
        return losefocus, gainfocus

    def set_focus(self):
        """Give focus to active widget."""
        if self.active_item:
            self.give_focus_to_widget(self.active_item.panel)
            self.active_item.set_statusbar_text()

    def configure_items_grid(self, panel, active_weight=None):
        """Adjust items panel grid row sizes after navigate add or delete."""
        if active_weight is None:
            active_weight = max(2, len(self.order) - 1)
        for index, item in enumerate(self.order):
            item.get_top_widget().grid(
                row=index, column=0, sticky=tkinter.NSEW
            )
            panel.grid_columnconfigure(0, weight=1, uniform="c")

            # next line may do as alternative to line above
            # panel.grid_columnconfigure(0, weight=1)

            if item is self.active_item:
                panel.grid_rowconfigure(
                    index, weight=active_weight, uniform="v"
                )
            else:
                panel.grid_rowconfigure(
                    index, weight=0 if item.ui.single_view else 1, uniform="v"
                )

    def object_display_count(self, key):
        """Return count of widgets which display object of key."""
        return self.object_panel_count.get(key)

    def set_insert_or_delete_on_all_items(self):
        """Convert edit display to insert display.

        PGN scores displayed for editing from a database are not closed if the
        database is closed.  They are converted to insert displays and can
        be used to add new records to databases opened later.

        """
        for item in self.stack:
            item.sourceobject = None

    def forget_payload(self, parent):
        """Do nothing: compatibility with instances of Display subclasses."""

    def insert_payload(self, parent):
        """Do nothing: compatibility with instances of Display subclasses."""

    def is_payload_available(self):
        """Return True if items displayed: implies active item exists."""
        return bool(self.stack)

    def bind_for_widget_without_focus(self):
        """Return True if this item has the focus about to be lost."""
        if self.active_item is None:
            return False
        if not self.active_item.has_focus():
            return False
        self.active_item.bind_for_widget_navigation()
        return True


class DisplayItemsStub:
    """Stub manager for set of displayed widgets.

    Tk Frames, each containing an item and usually called a panel in ChessTab,
    can be displayed in a container with the convention the active item is the
    [-1] element of self.stack.  self.panel_object_map maps panel to item.

    The item can be displayed in a tk Toplevel with no other items.  The
    is_mapped_panel method and active_item property return values meaning the
    item is always the active item and no other item can be available to be
    active.

    """

    def __init__(self):
        """Create control data structures for  widgets."""
        # Only stack and panel_object_map are referenced when module written.
        # self.order = ()
        self.stack = (None,)
        self.panel_object_map = frozenset()
        # self.object_panel_count = frozenset()

    @property
    def active_item(self):
        """Return the active item."""
        return self.stack[-1]

    def is_mapped_panel(self, panel):
        """Return True if panel is in self.panel_object_map."""
        return panel in self.panel_object_map



# chessgames_winedptmulti.py
# Copyright 2010 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import chess games when using DPT on Wine.

"chessgames_winedptmulti" is an obsolete version of "chessgames_winedptchunk".

This module has survived as an example of how to code multi-step in Python.
"""

if __name__ == "__main__":
    from . import APPLICATION_NAME

    APP_NAME = " ".join((APPLICATION_NAME, "(WineDPTMulti)"))
    try:
        from solentware_misc.gui.startstop import (
            start_application_exception,
            stop_application,
            application_exception,
        )
    except Exception as error:
        import tkinter.messagebox

        try:
            tkinter.messagebox.showerror(
                title="Start Exception",
                message=".\n\nThe reported exception is:\n\n".join(
                    (
                        "".join(
                            (
                                "Unable to import ",
                                "solentware_misc.gui.startstop module",
                            )
                        ),
                        str(error),
                    )
                ),
            )
        except tkinter.TclError:
            pass
        raise SystemExit(
            "Unable to import start application utilities"
        ) from error
    try:
        from .gui.chess import Chess
    except Exception as error:
        start_application_exception(error, appname=APP_NAME, action="import")
        raise SystemExit(" import ".join(("Unable to", APP_NAME))) from error
    try:
        app = Chess(allowcreate=True, dptmultistepdu=True)
    except Exception as error:
        start_application_exception(
            error, appname=APP_NAME, action="initialise"
        )
        raise SystemExit(
            " initialise ".join(("Unable to", APP_NAME))
        ) from error
    try:
        app.root.mainloop()
    except SystemExit:
        stop_application(app, app.root)
        raise
    except Exception as error:
        application_exception(
            error,
            app,
            app.root,
            title=APP_NAME,
            appname=APP_NAME,
        )
        raise SystemExit(
            " reporting exception in ".join(("Exception while", APP_NAME))
        ) from error



# gnudu_file.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with gnu.chessgnudu to database."""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..gnu.chessgnudu import ChessDatabase

    FileWidget(ChessDatabase, "dbm.gnu")



# chessgames_winedptchunk.py
# Copyright 2010 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import chess games when using DPT on Wine.

If games are not being imported, use sibling module "chessgames" instead.

Run "python -m chesstab.chessgames_winedptchunk" assuming chesstab is in
site-packages and Python3.3 or later is the system Python.

This module works around an 'out of memory' situation which arises when running
under Wine on FreeBSD using the DPT database engine to import games from a PGN
file containing more than about 5000 normal-size game scores on a system with
1Gb memory.

At least one version of Wine from before 2010 is not affected, and maybe some
other more recent versions are not affected.  It is not known whether other
systems capable of running Wine are affected.

"chessgames_winedptchunk" on Wine takes 12 times longer than "chessgames" on
Microsoft Windows, both using DPT, to complete an import of a large number of
games. Two million games is large: but 10,000 games is not large.

"""

if __name__ == "__main__":
    from . import APPLICATION_NAME

    APP_NAME = " ".join((APPLICATION_NAME, "(WineDPTChunk)"))
    try:
        from solentware_misc.gui.startstop import (
            start_application_exception,
            stop_application,
            application_exception,
        )
    except Exception as error:
        import tkinter.messagebox

        try:
            tkinter.messagebox.showerror(
                title="Start Exception",
                message=".\n\nThe reported exception is:\n\n".join(
                    (
                        "".join(
                            (
                                "Unable to import ",
                                "solentware_misc.gui.startstop module",
                            )
                        ),
                        str(error),
                    )
                ),
            )
        except tkinter.TclError:
            pass
        raise SystemExit(
            "Unable to import start application utilities"
        ) from error
    try:
        from .gui.chess import Chess
    except Exception as error:
        start_application_exception(error, appname=APP_NAME, action="import")
        raise SystemExit(" import ".join(("Unable to", APP_NAME))) from error
    try:
        app = Chess(allowcreate=True, dptchunksize=5000)
    except Exception as error:
        start_application_exception(
            error, appname=APP_NAME, action="initialise"
        )
        raise SystemExit(
            " initialise ".join(("Unable to", APP_NAME))
        ) from error
    try:
        app.root.mainloop()
    except SystemExit:
        stop_application(app, app.root)
        raise
    except Exception as error:
        application_exception(
            error,
            app,
            app.root,
            title=APP_NAME,
            appname=APP_NAME,
        )
        raise SystemExit(
            " reporting exception in ".join(("Exception while", APP_NAME))
        ) from error



# read_pgn.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Read a PGN file and calculate all indexing."""

from pgn_read.samples._utilities import main

from ..core.pgn import GameUpdate


if __name__ == "__main__":
    main(game_class=GameUpdate, samples_title="Sample ChessTab Update Report")



# chessapsw.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database using sqlite3."""

# from apsw import ConstraintError

from solentware_base import apsw_database
from solentware_base.core.constants import (
    FILEDESC,
)

from ..core.filespec import FileSpec
from ..basecore import database


class ChessDatabaseError(Exception):
    """Exception class for chessapsw module."""


class ChessDatabase(database.Database, apsw_database.Database):
    """Provide access to a database of games of chess."""

    _deferred_update_process = "chesstab.apsw.chessapswdu"

    def __init__(
        self,
        sqlite3file,
        use_specification_items=None,
        dpt_records=None,
        **kargs,
    ):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        names = FileSpec(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
        )

        if not kargs.get("allowcreate", False):
            try:
                for table_name in names:
                    if FILEDESC in names[table_name]:
                        del names[table_name][FILEDESC]
            except Exception as error:
                if __name__ == "__main__":
                    raise
                raise ChessDatabaseError(
                    "sqlite3 description invalid"
                ) from error

        try:
            super().__init__(
                names,
                sqlite3file,
                use_specification_items=use_specification_items,
                **kargs,
            )
        except ChessDatabaseError as error:
            if __name__ == "__main__":
                raise
            raise ChessDatabaseError("sqlite3 description invalid") from error

    def _delete_database_names(self):
        """Override and return tuple of filenames to delete."""
        return (self.database_file,)



# file_widget_fastreload.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Try fast load of a ChessTab database from fast unload output.

Purpose is to verify fast load functions correctly for at least one set
of 'TAPE' files: in particular a set generated by fast unload.

The attempt to write a fast load input files generator for PGN files from
the description of file formats in DPT DBA Guide failed: so see if there
is a way fast load works and copy that.
"""

import tkinter
import os
import time

import tkinter.messagebox
import tkinter.filedialog

import dptdb.dptapi

from solentware_base.core.constants import DPT_SYSFL_FOLDER

from ..dpt import chessdpt


def file_fastreload(dbpath, inputdir):
    """Open database, fast load Games file, and close database."""
    print(time.ctime())
    cdb = ChessDatabase(dbpath, allowcreate=True)
    cdb.open_database()
    for table in cdb.table.values():
        # Load accepts positional arguments only.
        # Want 'dir' argument as 'dbpath', not the default '#FASTIO' via
        # definition of FUNLOAD_DIR.
        # So have to specify options where FLOAD_DEFAULT is the default
        # option.
        table.opencontext.Load(dptdb.dptapi.FLOAD_DEFAULT, 0, None, inputdir)
    cdb.close_database()
    print(time.ctime())


class ChessDatabase(chessdpt.ChessDatabase):
    """Provide access to a database of games of chess."""

    def __init__(self, databasefolder, sysfolder=None, **kargs):
        """Extend to override default sysfolder."""
        if databasefolder:
            databasefolder = os.path.abspath(databasefolder)
            if sysfolder is None:
                sysfolder = os.path.join(databasefolder, DPT_SYSFL_FOLDER)
        super().__init__(databasefolder, sysfolder=sysfolder, **kargs)

    # Set default parameters for fastload as deferred update alternative.
    def create_default_parms(self):
        """Create default parms.ini file for fastload without transactions.

        This means transactions are disabled and a small number of DPT
        buffers.

        """
        if not os.path.exists(self.parms):
            parms = open(self.parms, "w")
            try:
                parms.write("RCVOPT=X'00' " + os.linesep)
                parms.write("MAXBUF=100 " + os.linesep)
            finally:
                parms.close()


class FileWidget:
    """Select ChessTab database to fast reload from fast unload output."""

    def __init__(self):
        """Import games into database using engine_name database engine."""
        root = tkinter.Tk()
        root.wm_title("Fast Reload ChessTab Database")
        root.wm_iconify()
        dbdir = tkinter.filedialog.askdirectory(
            title="Fast Reload ChessTab Database"
        )
        if dbdir:
            inputdir = tkinter.filedialog.askdirectory(
                title="Fast Reload Input Directory"
            )
            if inputdir:
                if tkinter.messagebox.askyesno(
                    title="Fast Reload Games",
                    message="Proceed with Fast Reload",
                ):
                    file_fastreload(dbdir, inputdir)
        root.destroy()


if __name__ == "__main__":
    FileWidget()



# cqldbdelete.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise delete toplevel to delete ChessQL statement record.

ChessQL statements obey the syntax published for CQL version 6.0.1 (by Gady
Costeff).

"""
import tkinter

from solentware_grid.gui.datadelete import DataDelete

from .cqltoplevel import CQLToplevel
from .topleveltext import DeleteText


class CQLDbDelete(DeleteText, DataDelete):
    """Delete ChessQL statement from database.

    parent is used as the master argument in a CQLToplevel call.

    ui is used as the ui argument in a CQLToplevel call.

    parent, oldobject, and the CQLToplevel instance created, are used as
    arguments in the super.__init__ call.

    Attribute text_name provides the name used in widget titles and message
    text.

    Methods _get_title_for_object and _set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    text_name = "ChessQL Statement"

    def __init__(self, parent=None, oldobject=None, ui=None):
        """Extend and create toplevel widget to delete ChessQL statement."""
        # Toplevel title set '' in __init__ and to proper value in _initialize.
        super().__init__(
            instance=oldobject,
            parent=parent,
            oldview=CQLToplevel(master=parent, ui=ui),
            title="",
        )
        self._initialize()

    def _get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a ChessQL statement object_.

        Default value of object_ is object attribute from DataDelete class.

        """
        if object_ is None:
            object_ = self.object
        return "  ".join(
            (
                self.text_name.join(("Delete ", ":")),
                object_.value.get_name_text(),
            )
        )

    @property
    def ui_base_table(self):
        """Return the User Interface TagRosterGrid object."""
        return self.ui.base_partials

    @property
    def ui_items_in_toplevels(self):
        """Return the User Interface objects in Toplevels."""
        return self.ui.partials_in_toplevels

    @property
    def ui(self):
        """Return the User Interface object from 'read-only' view."""
        return self.oldview.ui

    @staticmethod
    def _set_item(view, object_):
        """Populate view with the CQL query extracted from object_."""
        view.cql_statement.process_statement(object_.get_srvalue())
        view.set_and_tag_item_text()

    def dialog_ok(self):
        """Return delete action response (True for deleted).

        Delegate to superclass if the active item on the main display is not a
        'Position | Partial' widget and is not set to edit the same record.

        """
        if self.ui.partial_items.active_item:
            if self.ui.partial_items.active_item.sourceobject is None:
                tkinter.messagebox.showinfo(
                    parent=self.parent,
                    title=self._get_title_for_object(),
                    message="".join(
                        (
                            "Cannot use this delete dialogue while the ",
                            "active item in cql queries is one opened by ",
                            "menu action 'Position | Partial'.",
                        )
                    ),
                )
                return False
            if (
                self.ui.partial_items.active_item.sourceobject.key
                == self.object.key
            ):
                tkinter.messagebox.showinfo(
                    parent=self.parent,
                    title=self._get_title_for_object(),
                    message="".join(
                        (
                            "Cannot use this delete dialogue while the ",
                            "active item in cql queries is one that ",
                            "displays the record being deleted.",
                        )
                    ),
                )
                return False
        return super().dialog_ok()

    def delete(self, commit=True):
        """Delegate to superclass to delete record then delete game list."""
        if commit:
            self.datasource.dbhome.start_transaction()
        super().delete(commit=False)
        cqls = self.ui.partialpositionds(
            self.ui.base_games.datasource.dbhome,
            self.ui.base_games.datasource.dbset,
            self.ui.base_games.datasource.dbset,
            newrow=None,
        )
        assert self.object.newrecord is None
        cqls.forget_cql_statement_games(self.object, commit=False)
        if commit:
            self.datasource.dbhome.commit()



# sqlite3du_file.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with sqlite.chesssqlite3du to database."""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..sqlite.chesssqlite3du import ChessDatabase

    FileWidget(ChessDatabase, "sqlite3")



# chessdpt.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess games database using DPT database via dptdb.dptapi."""

import os
import tkinter.messagebox

# pylint will always give import-error message on non-Microsoft Windows
# systems.
# Wine counts as a Microsft Windows system.
# It is reasonable to not install 'dptdb.dptapi'.
# The importlib module is used to import chessdpt if needed.
from dptdb.dptapi import (
    FIFLAGS_FULL_TABLEB,
    FIFLAGS_FULL_TABLED,
)

from .. import APPLICATION_NAME
from . import chessdptnofistat


class ChessDatabase(chessdptnofistat.ChessDatabase):
    """Provide access to a database of games of chess.

    The open_database() method is extended in a number of ways, all but
    one with a new name.  These methods take the FISTAT flags into
    account when attempting to open the database.
    """

    def __init__(self, *args, **kwargs):
        """Define chess database.

        See superclass for argument descriptions.

        """
        super().__init__(*args, **kwargs)
        self._broken_sizes = {}

    def adjust_database_for_retry_import(self, files):
        """Increase file sizes taking file full into account."""
        # Increase the size of files allowing for the file full condition
        # which occurred while doing a deferred update for import.
        for dbn, broken_sizes in self._broken_sizes.items():
            self.table[dbn].increase_size_of_full_file(
                self.dbenv,
                self.table[dbn].get_file_parameters(self.dbenv),
                broken_sizes,
            )

    def open_database(self, files=None):
        """Return True if all files are opened in Normal mode (FISTAT == 0)."""
        super().open_database(files=files)
        fistat = {}
        for dbo in self.table.values():
            fistat[dbo] = dbo.get_file_parameters(self.dbenv)["FISTAT"]
        for dbo in self.table.values():
            if fistat[dbo][0] != 0:
                break
        else:
            self.increase_database_size(files=None)
            return True

        # At least one file is not in Normal state
        report = "\n".join(
            [
                "\t".join((os.path.basename(dbo.file), fistat[dbo][1]))
                for dbo in self.table.values()
            ]
        )
        tkinter.messagebox.showinfo(
            title="Open",
            message="".join(
                (
                    APPLICATION_NAME,
                    " has opened the database but some of the files are ",
                    "not in the Normal state.\n\n",
                    report,
                    "\n\n",
                    APPLICATION_NAME,
                    " will close the database on dismissing this ",
                    "dialogue.\n\nRestore the database from backups, or ",
                    "source data, before trying again.",
                )
            ),
        )
        self.close_database()

    def _delete_database_names(self):
        """Override and return tuple of filenames to delete."""
        names = [self.sysfolder]
        for value in self.table.values():
            names.append(value.file)
        return tuple(names)

    def get_archive_names(self, file=None):
        """Return names and operating system files for archives and guards."""
        names = [v.file for k, v in self.table.items() if k == file]
        archives = {}
        guards = {}
        for name in names:
            archiveguard = ".".join((name, "grd"))
            archivefile = ".".join((name, "bz2"))
            for box, arch in ((archives, archivefile), (guards, archiveguard)):
                if os.path.exists(arch):
                    box[name] = arch
        return (names, archives, guards)

    def open_after_import(self, files=()):
        """Return open context after doing database engine specific actions.

        For DPT clear the file sizes before import area if the database was
        opened successfully as there is no need to retry the import.

        """
        super().open_database()
        fistat = {}
        file_sizes_for_import = {}
        for dbn, dbo in self.table.items():
            gfp = dbo.get_file_parameters(self.dbenv)
            fistat[dbo] = gfp["FISTAT"]
            if dbn in files:
                file_sizes_for_import[dbn] = gfp
        for dbo in self.table.values():
            if fistat[dbo][0] != 0:
                break
        else:
            # Assume all is well as file status is 0
            # Or just do nothing (as file_sizes_for_import may be removed)
            self.increase_database_size(files=None)
            self.mark_partial_positions_to_be_recalculated()
            return True
        # At least one file is not in Normal state after Import.
        # Check the files that had imports applied
        for file_sizes in file_sizes_for_import.values():
            # pylint message unused variable.
            # Document what seemed to matter at some point.
            # status = file_sizes["FISTAT"][0]
            flags = file_sizes["FIFLAGS"]
            if not (
                (flags & FIFLAGS_FULL_TABLEB) or (flags & FIFLAGS_FULL_TABLED)
            ):
                break
        else:
            # The file states are consistent with the possibility that the
            # import failed because at least one file was too small.
            # The file size information is kept for calculating an increase
            # in file size before trying the import again.
            tkinter.messagebox.showinfo(
                title="Open",
                message="".join(
                    (
                        "The import failed.\n\n",
                        APPLICATION_NAME,
                        " has opened the database but some of the files are ",
                        "full.  The database may not be usable.",
                    )
                ),
            )
            # self.close_database()
            return None
        # At least one file is not in Normal state.
        # None of these files had deferred updates for Import or the state does
        # not imply a file full condition where deferred updates occured.
        report = "\n".join(
            [
                "\t".join((os.path.basename(dbo.file), fistat[dbo][1]))
                for dbo in self.table.values()
            ]
        )
        tkinter.messagebox.showinfo(
            title="Open",
            message="".join(
                (
                    APPLICATION_NAME,
                    " has opened the database but some of the files are ",
                    "not in the Normal state.\n\n",
                    report,
                    "\n\nAt least one of these files is neither just ",
                    "marked Deferred Update nor marked Full.  The ",
                    "database may not be usable.",
                )
            ),
        )
        # self.close_database()
        return True

    def save_broken_database_details(self, files=()):
        """Save database engine specific detail of broken files to be restored.

        It is assumed that the Database Services object exists.

        """
        self._broken_sizes.clear()
        broken = self._broken_sizes
        for file in files:
            broken[file] = self.table[file].get_file_parameters(self.dbenv)



# dptdumulti.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file using DPT multi-step deferred update."""


if __name__ == "__main__":
    import tkinter

    import tkinter.messagebox
    import tkinter.filedialog

    from chesstab.dpt.chessdptdumulti import chess_database_du

    root = tkinter.Tk()
    root.wm_title(string="Test Import Chess Games")
    root.wm_iconify()
    dbdir = tkinter.filedialog.askdirectory(title="Open Chess database folder")
    if dbdir:
        filename = tkinter.filedialog.askopenfilename(
            title="PGN file of Games",
            defaultextension=".pgn",
            filetypes=(("PGN Chess Games", "*.pgn"),),
        )
        if filename:
            if tkinter.messagebox.askyesno(
                title="Import Games", message="Proceed with import"
            ):
                chess_database_du(dbdir, (filename,), {})
    root.destroy()



# cql_gamelist_query.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""CQLGamelistQuery class for methods shared by most *ListGrid classes.

'Most' means CQLListGrid, GameListGrid, and QueryListGrid.

See allgrid.AllGrid class for methods identical in EngineListGrid class too.
"""
import tkinter

from ..gui.eventspec import EventSpec


class CQLGameListQuery:
    """Provide methods shared by CQLListGrid, GameListGrid, and QueryGrid."""

    def _set_popup_bindings(self, popup, bindings=()):
        """Set popup menu bindings for game list grid."""
        for accelerator, function in bindings:
            popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, popup),
                accelerator=accelerator[2],
            )

    def _add_cascade_menu_to_popup(self, index, popup, bindings=None):
        """Add cascade_menu, and bindings, to popup if not already present.

        The index is used as the label on the popup menu when visible.

        The bindings are not applied if cascade_menu is alreay in popup menu.

        """
        # Cannot see a way of asking 'Does entry exist?' other than:
        try:
            popup.index(index)
        except tkinter.TclError:
            cascade_menu = tkinter.Menu(master=popup, tearoff=False)
            popup.add_cascade(label=index, menu=cascade_menu)
            if bindings is None:
                return
            self._set_popup_bindings(cascade_menu, bindings)

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order."""
        del event
        self.ui.give_focus_backward(self)
        return "break"

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order."""
        del event
        self.ui.give_focus_forward(self)
        return "break"

    @staticmethod
    def traverse_round(event=None):
        """Give focus to next widget within active item in traversal order."""
        del event
        return "break"

    def set_focus(self):
        """Give focus to this widget."""
        self.frame.focus_set()
        if self.ui.single_view:
            self.ui.show_just_panedwindow_with_focus(self.frame)

    def focus_set_frame(self, event=None):
        """Adjust widget which is losing focus then delegate to superclass."""
        self.ui.set_bindings_on_item_losing_focus_by_pointer_click()
        super().focus_set_frame(event=event)

    def bind_for_widget_without_focus(self):
        """Return True if this item has the focus about to be lost."""
        if self.get_frame().focus_displayof() != self.get_frame():
            return False

        # Nothing to do on losing focus.
        return True

    def get_top_widget(self):
        """Return topmost widget for game display.

        The topmost widget is put in a container widget in some way.

        """
        # Superclass DataGrid.get_frame() method returns the relevant widget.
        # Name, get_top_widget, is compatible with Game and Partial names.
        return self.get_frame()

    def get_visible_selected_key(self):
        """Return selected key if it is visible and display dialogue if not.

        Getting the key is delegated to superclass.

        """
        key = super().get_visible_selected_key()
        if key is None:
            tkinter.messagebox.showinfo(
                parent=self.parent,
                title="Display Item",
                message="No record selected or selected record is not visible",
            )
        return key

    def _display_selected_item(self, key):
        """Create display item for selected record."""
        selected = self.get_visible_record(key)
        if selected is not None:
            try:
                self._display_selected_item_kind(key, selected)

            # Bad data, or perhaps legitimate absence of data, in a record
            # leads first to a SyntaxError in ast.literal_eval() and then
            # to a StopIteration condition.  With the DPT database engine
            # absence of data is legitimate but not expected.  In the other
            # database engines bad data is expected to be the cause.
            except StopIteration:
                if len(selected.srvalue) < 10:
                    message = selected.srvalue
                else:
                    message = "...".join(
                        (
                            selected.srvalue[:5],
                            selected.srvalue[-5:],
                        )
                    )
                tkinter.messagebox.showinfo(
                    parent=self.parent,
                    title="Display Item",
                    message=message.join(
                        (
                            "Unable to build display item from ",
                            " data",
                        )
                    ),
                )

    def select_down(self):
        """Extend to show selection summary in status bar."""
        super().select_down()
        self.set_selection_text()

    def select_up(self):
        """Extend to show selection summary in status bar."""
        super().select_up()
        self.set_selection_text()

    def cancel_selection(self):
        """Extend to clear selection summary from status bar."""
        if self.selection:
            self.ui.statusbar.set_status_text("")
        super().cancel_selection()

    def _set_background_on_display_row_under_pointer(self, key):
        self.objects[key].set_background_on_display(self.get_row_widgets(key))
        self.set_row_under_pointer_background(key)

    def _set_background_normal_row_under_pointer(self, key):
        self.objects[key].set_background_normal(self.get_row_widgets(key))
        self.set_row_under_pointer_background(key)

    def _configure_frame_and_initial_event_bindings(self):
        self._configure_frame()
        self._set_event_bindings_frame(
            (
                (EventSpec.tab_traverse_forward, self.traverse_forward),
                (EventSpec.tab_traverse_backward, self.traverse_backward),
                (EventSpec.tab_traverse_round, self.traverse_round),
                # Remove entries when binding implemented in solentware_grid.
                (
                    EventSpec.score_enable_F10_popupmenu_at_top_left,
                    self.show_grid_or_row_popup_menu_at_top_left_by_keypress,
                ),
                (
                    EventSpec.score_enable_F10_popupmenu_at_pointer,
                    self.show_grid_or_row_popup_menu_at_pointer_by_keypress,
                ),
            )
        )



# board.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Classes to draw positions on chess boards.

The Board class is used to show positions in a game of chess.

Fonts such as Chess Merida, if installed, are used to represent pieces on the
board.

Character equivalents for the pieces if none of these fonts are installed:

l  King
w  Queen
t  Rook
v  Bishop
m  Knight
o  Pawn
"""

import tkinter
import tkinter.font

from solentware_bind.gui.bindings import Bindings

from pgn_read.core.constants import (
    FEN_WHITE_KING,
    FEN_WHITE_QUEEN,
    FEN_WHITE_ROOK,
    FEN_WHITE_BISHOP,
    FEN_WHITE_KNIGHT,
    FEN_WHITE_PAWN,
    FEN_BLACK_KING,
    FEN_BLACK_QUEEN,
    FEN_BLACK_ROOK,
    FEN_BLACK_BISHOP,
    FEN_BLACK_KNIGHT,
    FEN_BLACK_PAWN,
    FEN_WHITE_PIECES,
    FEN_BLACK_PIECES,
)
from pgn_read.core.squares import Squares

from . import constants
from ..core.constants import NOPIECE

_pieces = (
    NOPIECE,
    FEN_WHITE_KING,
    FEN_WHITE_QUEEN,
    FEN_WHITE_ROOK,
    FEN_WHITE_BISHOP,
    FEN_WHITE_KNIGHT,
    FEN_WHITE_PAWN,
    FEN_BLACK_KING,
    FEN_BLACK_QUEEN,
    FEN_BLACK_ROOK,
    FEN_BLACK_BISHOP,
    FEN_BLACK_KNIGHT,
    FEN_BLACK_PAWN,
)
# Chess pieces are solid unicode piece characters or the black pieces on
# light square in the four Chess fonts, Cases, Lucena, Merida and Motif,
# by Armando H Marroquin.
# The four 'Chess *' fonts were downloaded from
# www.enpassant.dk/chess/fonteng.htm
# Which set of chess pieces are used depends on the chosen font when the
# board is drawn: sometimes the characters "l", "w", "t", "v", "m", and
# "o", will appear: only to be replaced by the next action which causes
# the board to be redrawn.
_piece_chars = (
    dict(
        zip(
            _pieces,
            (
                "",
                "l",
                "w",
                "t",
                "v",
                "m",
                "o",
                "l",
                "w",
                "t",
                "v",
                "m",
                "o",
            ),
        )
    ),  # 'Chess Cases', 'Chess Lucena', 'Chess Merida', and 'ChessMotif'.
    dict(
        zip(
            _pieces,
            (
                "",
                "\u265a",
                "\u265b",
                "\u265c",
                "\u265d",
                "\u265e",
                "\u265f",
                "\u265a",
                "\u265b",
                "\u265c",
                "\u265d",
                "\u265e",
                "\u265f",
            ),
        )
    ),  # Both solid unicode chess pieces.
    dict(
        zip(
            _pieces,
            (
                "",
                "\u265a",
                "\u265b",
                "\u265c",
                "\u265d",
                "\u265e",
                "\u265f",
                "\u2654",
                "\u2655",
                "\u2656",
                "\u2657",
                "\u2658",
                "\u2659",
            ),
        )
    ),  # White solid and black outline unicode chess pieces.
    dict(
        zip(
            _pieces,
            (
                "",
                "\u2654",
                "\u2655",
                "\u2656",
                "\u2657",
                "\u2658",
                "\u2659",
                "\u265a",
                "\u265b",
                "\u265c",
                "\u265d",
                "\u265e",
                "\u265f",
            ),
        )
    ),  # White outline and black solid unicode chess pieces.
    dict(
        zip(
            _pieces,
            (
                "",
                "\u2654",
                "\u2655",
                "\u2656",
                "\u2657",
                "\u2658",
                "\u2659",
                "\u2654",
                "\u2655",
                "\u2656",
                "\u2657",
                "\u2658",
                "\u2659",
            ),
        )
    ),  # Both outline unicode chess pieces.
)
del _pieces


class Board(Bindings):
    """Chess board widget.

    Frame containing an 8x8 grid of Text widgets representing chess board
    with a font used to denote the pieces.

    """

    litecolor = constants.LITECOLOR
    darkcolor = constants.DARKCOLOR
    whitecolor = constants.WHITECOLOR
    blackcolor = constants.BLACKCOLOR
    boardfont = constants.PIECES_ON_BOARD_FONT
    _preferred_pieces = frozenset(constants.PREFERRED_PIECES)

    def __init__(self, master, boardborder=2, boardfont=None, ui=None):
        """Create board widget.

        The container catches application resizing and reconfigures
        Board to it's new size. The board then processes the Canvases to
        adjust fonts. Neither propagates geometry changes to it's master.

        """
        super().__init__()
        self.ui = ui
        if boardfont:
            self.boardfont = boardfont

        self.squares = {}
        self.boardsquares = {}
        # self.boardfont is name of named font or a font instance
        try:
            self.font = tkinter.font.nametofont(self.boardfont).copy()
        except (AttributeError, tkinter.TclError):
            self.font = self.boardfont.copy()
        self.container = tkinter.Frame(
            master=master, cnf=dict(width=0, height=0)
        )
        self.bind(
            self.container,
            "<Configure>",
            function=self.try_event(self._on_configure_container),
        )
        self.board = tkinter.Frame(
            master=self.container,
            cnf=dict(borderwidth=boardborder, relief=tkinter.SUNKEN),
        )
        board = self.board
        boardsquares = self.boardsquares
        font = self.font
        litecolor = self.litecolor
        darkcolor = self.darkcolor
        board.pack(anchor=tkinter.W)
        board.grid_propagate(False)
        for index in range(8):
            board.grid_rowconfigure(index, weight=1, uniform="r")
            board.grid_columnconfigure(index, weight=1, uniform="c")
            for file_index in range(8):
                square = index * 8 + file_index
                if (index + file_index) % 2 == 0:
                    scolor = litecolor
                else:
                    scolor = darkcolor
                boardsquares[square] = tkinter.Label(
                    board, font=font, background=scolor
                )
                boardsquares[square].grid(
                    column=file_index, row=index, sticky=tkinter.NSEW
                )

    def _configure_font_size(self, side):
        """Adjust font size after container widget has been resized."""
        self.font.configure(size=-(side * 3) // 32)

    def _on_configure_container(self, event=None):
        """Reconfigure board after container widget has been resized."""
        del event
        side = min(self.container.winfo_width(), self.container.winfo_height())
        self.board.configure(width=side, height=side)
        self._configure_font_size(side)
        self.draw_board()

    def draw_board(self):
        """Set font size to match board size and redraw pieces."""
        if self.font.actual()["family"] in self._preferred_pieces:
            piece_chars = _piece_chars[0]
        else:
            piece_chars = _piece_chars[1]
        for index, piece in self.squares.items():
            if piece in FEN_WHITE_PIECES:
                pcolor = self.whitecolor
            elif piece in FEN_BLACK_PIECES:
                pcolor = self.blackcolor
            elif piece == NOPIECE:
                if index % 2 == 0:
                    pcolor = self.darkcolor
                else:
                    pcolor = self.litecolor
            else:
                continue
            self.boardsquares[index].configure(
                foreground=pcolor, text=piece_chars[piece]
            )

    def get_top_widget(self):
        """Return top level frame of this widget."""
        return self.container

    def set_color_scheme(self):
        """Set background color for Canvas for each square."""
        for rank_index in range(8):
            for file_index in range(8):
                square = rank_index * 8 + file_index
                if (rank_index + file_index) % 2 == 0:
                    scolor = self.darkcolor
                else:
                    scolor = self.litecolor
                self.boardsquares[square].configure(background=scolor)

    def set_board(self, board):
        """Redraw widget to display the new position in board.

        board is a list of pieces where element index maps to square.

        """
        squares = self.squares
        occupied = list(squares.keys())
        squares.clear()
        square_properties = Squares.squares
        for square, piece in board.items():
            squares[square_properties[square].number] = piece.name
        for square in occupied:
            if square not in squares:
                squares[square] = NOPIECE
        self.draw_board()
        for square in occupied:
            if squares[square] == NOPIECE:
                del squares[square]



# berkeleydbdu_file.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with berkeleydb.chessdbdu to database."""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..berkeleydb.chessberkeleydbdu import ChessDatabase

    FileWidget(ChessDatabase, "berkeleydb")
